#!/usr/bin/env python3

# [1] https://doi.org/10.1063/1.450106
#     Efficient recursive computation of molecular integrals over Cartesian
#     Gaussian functions
#     Obara, Saika, 1986
# [2] https://doi.org/10.1002/9781119019572
#     Molecular Electronic-Structure Theory
#     Helgaker, Jørgensen, Olsen
# [3] https://doi.org/10.1021/acs.jctc.7b00788
#     LIBRETA: Computerized Optimization and Code Synthesis for
#     Electron Repulsion Integral Evaluation
#     Jun Zhang
# [4] https://doi.org/10.1039/B413539C
#     Efficient evaluation of three-center two-electron integrals over Gaussian functions
#     Ahlrichs, 2004
# [5] EVALUATING MANY-ELECTRON MOLECULAR INTEGRALS FOR QUANTUM CHEMISTRY
#     James Christopher Womack, PhD Thesis
# [6] https://doi.org/10.1063/1.4983393
#     Efficient evaluation of three-center Coulomb integrals
#     Samu, Kállay, 2017
# [7] https://arxiv.org/pdf/2210.03192.pdf
#     Memory-Efficient Recursive Evaluation of 3-Center Gaussian Integrals
#     Asadchev, Valeev, 2022


import argparse
from datetime import datetime
import functools
import itertools as it
import os
from pathlib import Path
import random
import string
import sys
import textwrap
import time

from jinja2 import Template
import numpy as np
from sympy import (
    Array,
    cse,
    exp,
    Expr,
    flatten,
    Function,
    IndexedBase,
    Matrix,
    permutedims,
    pi,
    sqrt,
    Symbol,
    symbols,
    tensorcontraction as tc,
    tensorproduct as tp,
)


from sympy.codegen.ast import Assignment
from sympy.printing.numpy import NumPyPrinter
from sympy.printing.c import C99CodePrinter

# from pysisyphus.wavefunction.cart2sph import cart2sph_coeffs

try:
    from pysisyphus.config import L_MAX, L_AUX_MAX
except ModuleNotFoundError:
    L_MAX = 4
    L_AUX_MAX = 5


L_MAP = {
    0: "s",
    1: "p",
    2: "d",
    3: "f",
    4: "g",
    5: "h",
}

KEYS = (
    "cgto",
    "ovlp",
    "dpm",
    "dqpm",
    "qpm",
    "kin",
    "coul",
    "3c2e",
    "3c2e_sph",
)
ONE_THRESH = 1e-14


def make_py_func(repls, reduced, args=None, name=None, doc_str="", return_array=True):
    if args is None:
        args = list()
    # Generate random name, if no name was supplied
    if name is None:
        name = "func_" + "".join(
            [random.choice(string.ascii_letters) for i in range(8)]
        )

    # This allows using the 'boys' function without producing an error
    print_settings = {
        "allow_unknown_functions": True,
    }
    print_func = NumPyPrinter(print_settings).doprint
    assignments = [Assignment(lhs, rhs) for lhs, rhs in repls]
    py_lines = [print_func(as_) for as_ in assignments]
    return_val = print_func(reduced)
    try:
        n_return_vals = len(reduced)
    except TypeError:
        n_return_vals = 1

    tpl = Template(
        """
    def {{ name }}({{ args }}):
        {% if doc_str %}
        \"\"\"{{ doc_str }}\"\"\"
        {% endif %}

        {% for line in py_lines %}
        {{ line }}
        {% endfor %}

        {% if return_array %}
        # {{ n_return_vals }} item(s)
        return numpy.array({{ return_val }})
        {% else %}
        return {{ return_val }}
        {% endif %}
    """,
        trim_blocks=True,
        lstrip_blocks=True,
    )

    rendered = textwrap.dedent(
        tpl.render(
            name=name,
            args=args,
            py_lines=py_lines,
            return_val=return_val,
            n_return_vals=n_return_vals,
            return_array=return_array,
            doc_str=doc_str,
        )
    ).strip()
    return rendered


def make_c_func(repls, reduced, args=None, name=None, doc_str=""):
    if args is None:
        args = list()
    # Generate random name, if no name was supplied
    if name is None:
        name = "func_" + "".join(
            [random.choice(string.ascii_letters) for i in range(8)]
        )
    arg_strs = list()
    for arg in args:
        if arg.islower():
            arg_str = f"double {arg}"
        elif arg.isupper():
            arg_str = f"double {arg}[3]"
        else:
            raise Exception
        arg_strs.append(arg_str)
    args_str = ", ".join(arg_strs)

    # This allows using the 'boys' function without producing an error
    print_settings = {
        "allow_unknown_functions": True,
        # Without disabling contract some expressions will raise ValueError.
        "contract": False,
    }
    print_func = C99CodePrinter(print_settings).doprint
    assignments = [Assignment(lhs, rhs) for lhs, rhs in repls]
    repl_lines = [print_func(as_) for as_ in assignments]
    res_lines = [print_func(red) for red in reduced]
    res_len = len(reduced)

    signature = f"double * {name}({args_str})"

    tpl = Template(
        """
    {{ signature }} {
        {% if doc_str %}
        /* {{ doc_str }} */
        {% endif %}

        static double {{ res_name }}[{{ res_len}}];

        {% for line in repl_lines %}
        const double {{ line }}
        {% endfor %}

        {% for rhs in res_lines %}
        {{ res_name }}[{{ loop.index0}}] = {{ rhs }};
        {% endfor %}

        return {{ res_name }};
    }
    """,
        trim_blocks=True,
        lstrip_blocks=True,
    )

    rendered = textwrap.dedent(
        tpl.render(
            signature=signature,
            res_name="results",
            res_len=res_len,
            repl_lines=repl_lines,
            res_lines=res_lines,  # c_lines=c_lines,
            reduced=reduced,
            doc_str=doc_str,
            args_str=args_str,
        )
    ).strip()
    return rendered, signature


def canonical_order(L):
    inds = list()
    for i in range(L + 1):
        l = L - i
        for n in range(i + 1):
            m = i - n
            inds.append((l, m, n))
    return inds


def shell_iter(Ls):
    """Iterator over cartesian product of L values in Ls."""
    return it.product(*[canonical_order(L) for L in Ls])


class CartGTO3d(Function):
    """3D Cartesian Gaussian function; not normalized."""

    @classmethod
    @functools.cache
    def eval(cls, i, j, k, a, Xa, Ya, Za):
        Xa2 = Xa**2
        Ya2 = Ya**2
        Za2 = Za**2
        return (Xa**i) * (Ya**j) * (Za**k) * exp(-a * (Xa2 + Ya2 + Za2))


class CartGTOShell(Function):
    @classmethod
    def eval(cls, La_tot, a, Xa, Ya, Za):
        exprs = [CartGTO3d(*La, a, Xa, Ya, Za) for La, in shell_iter((La_tot,))]
        # print(CartGTO3d.eval.cache_info())
        return exprs


class TwoCenter1d(Expr):
    def __new__(self, a, A, b, B, C=None):
        self.a = a
        self.A = A
        self.b = b
        self.B = B
        self.C = C
        return super().__new__(self, a, A, b, B, C)

    @property
    def p(self):
        """Total exponent p."""
        return self.a + self.b

    @property
    def mu(self):
        """Reduced exponent mu."""
        return self.a * self.b / self.p

    @property
    def AB(self):
        """Relative coordinate/Gaussian separation X_AB."""
        return self.A - self.B

    @property
    def P(self):
        """Center-of-charge coordinate."""
        return (self.a * self.A + self.b * self.B) / self.p

    @property
    def PA(self):
        """Relative coordinate/Gaussian separation X_PA."""
        return self.P - self.A

    @property
    def PB(self):
        """Relative coordinate/Gaussian separation X_PB."""
        return self.P - self.B

    @property
    def PC(self):
        """Relative coordinate/Gaussian separation X_PC."""
        return self.P - self.C

    @property
    def K(self):
        return exp(-self.mu * self.AB**2)


class Multipole1d(TwoCenter1d):
    """1d multipole-moment integral of order 'e', between primitive 1d Gaussians
    Ga = G_i(a, r, A) and Gb = G_j(b, r, B) with Cartesian quantum number i and j,
    exponents a and b, centered at A (B). The origin of the multipole expansion is
    at C.
    """

    @functools.cache
    def eval(self, i, j, e):
        ang_moms = (i, j, e)
        if any([_ < 0 for _ in ang_moms]):
            return 0

        recur = self.eval

        def vrr(i, j, e, X):
            return X * recur(i, j, e) + 1 / (2 * self.p) * (
                i * recur(i - 1, j, e) + j * recur(i, j - 1, e) + e * recur(i, j, e - 1)
            )

        # Base case
        if all([_ == 0 for _ in ang_moms]):
            return sqrt(pi / self.p) * self.K
        # Decrement i
        elif i.is_positive:
            return vrr(i - 1, j, e, self.PA)
        # Decrement j
        elif j.is_positive:
            return vrr(i, j - 1, e, self.PB)
        elif e.is_positive:
            return vrr(i, j, e - 1, self.PC)


class Multipole3d(Function):
    @classmethod
    def eval(cls, La, Lb, a, b, A, B, Le, C):
        x, y, z = [
            Multipole1d(a, A[i], b, B[i], C[i]).eval(La[i], Lb[i], Le[i])
            for i in range(3)
        ]
        return x * y * z


class Multipole3dShell(Function):
    @classmethod
    def eval(cls, La_tot, Lb_tot, a, b, A, B, Le_tot=0, C=(0.0, 0.0, 0.0)):
        exprs = [
            Multipole3d(La, Lb, a, b, A, B, Le, C)
            for Le, La, Lb in shell_iter((Le_tot, La_tot, Lb_tot))
        ]
        # print(Multipole1d.eval.cache_info())
        return exprs


class DiagQuadrupole3dShell(Function):
    @classmethod
    def eval(cls, La_tot, Lb_tot, a, b, A, B, C=(0.0, 0.0, 0.0)):
        exprs = list()
        for Le in ((2, 0, 0), (0, 2, 0), (0, 0, 2)):
            for La, Lb in shell_iter((La_tot, Lb_tot)):
                exprs.append(Multipole3d(La, Lb, a, b, A, B, Le, C))
        # print(DiagQuadrupole3dShell.eval.cache_info())
        return exprs


class Overlap1d(TwoCenter1d):
    def eval(self, i, j):
        return Multipole1d(self.a, self.A, self.b, self.B).eval(i, j, 0)


class Overlap3dShell(Function):
    """Whole shell of 3d overlap integrals for a given L pair
    over primitive gaussians."""

    @classmethod
    def eval(cls, La_tot, Lb_tot, a, b, A, B):
        exprs = Multipole3dShell(La_tot, Lb_tot, a, b, A, B)
        # print(Multipole1d.eval.cache_info())
        return exprs


class Kinetic1d(TwoCenter1d):
    @functools.cache
    def eval(self, i, j):
        if i < 0 or j < 0:
            return 0

        recur = self.eval

        def recur_rel(i, j, X):
            return X * recur(i, j) + 1 / (2 * self.p) * (
                i * recur(i - 1, j) + j * recur(i, j - 1)
            )

        def recur_ovlp(i, j):
            return Overlap1d(self.a, self.A, self.b, self.B).eval(i, j)

        # Base case
        if i == 0 and j == 0:
            return (
                self.a - 2 * self.a**2 * (self.PA**2 + 1 / (2 * self.p))
            ) * recur_ovlp(i, j)
        # Decrement i
        elif i > 0:
            # Eq. (9.3.41)
            return recur_rel(i - 1, j, self.PA) + self.b / self.p * (
                2 * self.a * recur_ovlp(i, j) - i * recur_ovlp(i - 2, j)
            )
        # Decrement j
        elif j > 0:
            # Eq. (9.3.41)
            return recur_rel(i, j - 1, self.PB) + self.a / self.p * (
                2 * self.b * recur_ovlp(i, j) - j * recur_ovlp(i, j - 2)
            )


class Kinetic3d(Function):
    @classmethod
    def eval(cls, La, Lb, a, b, A, B):
        Tx, Ty, Tz = [Kinetic1d(a, A[i], b, B[i]).eval(La[i], Lb[i]) for i in range(3)]
        Sx, Sy, Sz = [Overlap1d(a, A[i], b, B[i]).eval(La[i], Lb[i]) for i in range(3)]
        return Tx * Sy * Sz + Sx * Ty * Sz + Sx * Sy * Tz


class Kinetic3dShell(Function):
    @classmethod
    def eval(cls, La_tot, Lb_tot, a, b, A, B):
        exprs = [
            Kinetic3d(La, Lb, a, b, A, B) for La, Lb in shell_iter((La_tot, Lb_tot))
        ]
        return exprs


# Placeholder for the Boys-function. The actual Boys-function will be
# imported in the generated python module.
boys = Function("boys")


class Coulomb(TwoCenter1d):
    """Nucleus at C."""

    @functools.cache
    def eval(self, i, k, m, j, l, n, N):
        ang_moms = (i, k, m, j, l, n)
        if any([am < 0 for am in ang_moms]):
            return 0

        def recur(N, *inds):
            """Simple wrapper to pass all required arguments."""
            return self.eval(*inds, N)  # , a, b, A, B, C)

        def decr(to_decr, decr_ind):
            one = np.zeros(3, dtype=int)
            one[decr_ind] = 1

            def Ni(inds):
                return inds[decr_ind]

            bra = np.array((i, k, m), dtype=int)
            ket = np.array((j, l, n), dtype=int)

            if to_decr == "bra":
                X = self.PA
                bra[decr_ind] -= 1
            else:
                X = self.PB
                ket[decr_ind] -= 1

            bra_decr = bra - one
            ket_decr = ket - one

            return (
                X[decr_ind] * recur(N, *bra, *ket)
                - self.PC[decr_ind] * recur(N + 1, *bra, *ket)
                + 1
                / (2 * self.p)
                * Ni(bra)
                * (recur(N, *bra_decr, *ket) - recur(N + 1, *bra_decr, *ket))
                + 1
                / (2 * self.p)
                * Ni(ket)
                * (recur(N, *bra, *ket_decr) - recur(N + 1, *bra, *ket_decr))
            )

        # Base case
        if all([am == 0 for am in ang_moms]):
            K = exp(-self.mu * self.AB.dot(self.AB))
            return 2 * pi / self.p * K * boys(N, self.p * self.PC.dot(self.PC))
        elif i > 0:
            return decr("bra", 0)
        elif j > 0:
            return decr("ket", 0)
        elif k > 0:
            return decr("bra", 1)
        elif l > 0:
            return decr("ket", 1)
        elif m > 0:
            return decr("bra", 2)
        elif n > 0:
            return decr("ket", 2)


class CoulombShell(Function):
    @classmethod
    def eval(cls, La_tot, Lb_tot, a, b, A, B, C=(0.0, 0.0, 0.0)):
        exprs = [
            # Coulomb(*La, *Lb, 0, a, b, A, B, C)
            Coulomb(a, A, b, B, C).eval(*La, *Lb, 0)
            for La, Lb in shell_iter((La_tot, Lb_tot))
        ]
        # print(Coulomb.eval.cache_info())
        return exprs


class TwoCenterTwoElectron(Function):
    @classmethod
    @functools.cache
    def eval(cls, ia, ja, ka, ib, jb, kb, N, a, b, A, B):
        ang_moms = np.array((ia, ja, ka, ib, jb, kb), dtype=int)
        ang_moms2d = ang_moms.reshape(-1, 3)
        if any([am < 0 for am in ang_moms]):
            return 0

        p = a + b
        P = (a * A + b * B) / p
        mu = (a * b) / p

        def recur(N, *inds):
            return cls(*inds, N, a, b, A, B)

        def vrr(bra_or_ket, cart_ind):
            assert bra_or_ket in ("bra", "ket")

            if bra_or_ket == "bra":
                ind1 = 0
                ind2 = 1
            else:
                ind1 = 1
                ind2 = 0
            exps = (a, b)
            X = (P - A, P - B)[ind1][cart_ind]

            l1 = ang_moms2d[ind1, cart_ind] - 1
            exp1 = exps[ind1]
            decr1 = ang_moms2d.copy()
            decr1[ind1, cart_ind] -= 1
            decr11 = decr1.copy()
            decr11[ind1, cart_ind] -= 1

            l2 = ang_moms2d[ind2, cart_ind]
            exp2 = exps[ind2]
            decr12 = ang_moms2d.copy()
            decr12[ind1, cart_ind] -= 1
            decr12[ind2, cart_ind] -= 1

            decr1 = decr1.flatten()
            decr11 = decr11.flatten()
            decr12 = decr12.flatten()

            return (
                X * recur(N + 1, *decr1)
                + l1
                / (2 * exp1)
                * (recur(N, *decr11) - exp2 / p * recur(N + 1, *decr11))
                + l2 / (2 * p) * recur(N + 1, *decr12)
            )

        # vrr_bra = functools.partial(vrr, bra_or_ket="bra")
        vrr_bra = functools.partial(vrr, "bra")
        vrr_ket = functools.partial(vrr, "ket")

        # Base case
        if (ang_moms == 0).all():
            AB = A - B
            return 2 * pi**2.5 / sqrt(p) / (a * b) * boys(N, mu * AB.dot(AB))
        elif ia > 0:
            return vrr_bra(0)
        elif ja > 0:
            return vrr_bra(1)
        elif ka > 0:
            return vrr_bra(2)
        elif ib > 0:
            return vrr_ket(0)
        elif jb > 0:
            return vrr_ket(1)
        elif kb > 0:
            return vrr_ket(2)


class TwoCenterTwoElectronShell(Function):
    @classmethod
    def eval(cls, La_tot, Lb_tot, a, b, A, B):
        exprs = [
            TwoCenterTwoElectron(*La, *Lb, 0, a, b, A, B)
            for La, Lb in shell_iter((La_tot, Lb_tot))
        ]
        # print(TwoCenterTwoElectron.eval.cache_info())
        return exprs


class ThreeCenterTwoElectronBase(Function):
    """
    https://pubs.rsc.org/en/content/articlelanding/2004/CP/b413539c

    There is an error in the base case (00|0). One must divide by
    sqrt(eta + gamma), not multiply.
    """

    @classmethod
    @functools.cache
    def eval(cls, ia, ja, ka, ib, jb, kb, ic, jc, kc, N, a, b, c, A, B, C):
        ang_moms = np.array((ia, ja, ka, ib, jb, kb, ic, jc, kc), dtype=int)
        ang_moms2d = ang_moms.reshape(-1, 3)
        if any([am < 0 for am in ang_moms]):
            return 0

        p = a + b
        P = (a * A + b * B) / p
        mu = (a * b) / p
        X_PC = P - C

        rho = p * c / (p + c)

        def recur(N, *inds):
            """Simple wrapper to pass all required arguments.

            Here we don't use ThreeCenterTwoElectronBase, but the derived classes,
            that provide the 'aux_vrr' attribute, to distinguish between the different
            vertical recursion relations."""
            return cls(*inds, N, a, b, c, A, B, C)

        def recur_hrr(cart_ind):
            """Horizontal recursion relation to transfer angular momentum in bra.

            (a, b+1_i|c) = (a + 1_i,b|c) + X_AB (ab|c)
            """
            assert N == 0
            incr_ang_moms = ang_moms2d.copy()
            incr_ang_moms[0, cart_ind] += 1  # Increase in left orbital
            incr_ang_moms[1, cart_ind] -= 1  # Decrease in right orbital

            decr_ang_moms = ang_moms2d.copy()
            decr_ang_moms[1, cart_ind] -= 1  # Decrease in right orbital

            incr_ang_moms = incr_ang_moms.flatten()
            decr_ang_moms = decr_ang_moms.flatten()

            AB_dir = (A - B)[cart_ind]

            return recur(N, *incr_ang_moms) + AB_dir * recur(N, *decr_ang_moms)

        def recur_vrr(cart_ind):
            assert (ib, jb, kb) == (0, 0, 0)
            assert (ic, jc, kc) == (0, 0, 0)

            decr_a = ang_moms2d.copy()
            decr_a[0, cart_ind] -= 1  # Decrease in bra left orbital
            decr_aa = decr_a.copy()
            decr_aa[0, cart_ind] -= 1  # Decrease in bra left orbital again

            PA_dir = (P - A)[cart_ind]
            PC_dir = (P - C)[cart_ind]
            ai = (ia, ja, ka)[cart_ind] - 1
            _2p = 2 * p

            decr_a = decr_a.flatten()
            decr_aa = decr_aa.flatten()

            return (
                PA_dir * recur(N, *decr_a)
                - rho / p * PC_dir * recur(N + 1, *decr_a)
                + ai / _2p * (recur(N, *decr_aa) - rho / p * recur(N + 1, *decr_aa))
            )

        def recur_vrr_aux(cart_ind):
            decr_c = ang_moms2d.copy()
            decr_c[2, cart_ind] -= 1

            decr_cc = decr_c.copy()
            decr_cc[2, cart_ind] -= 1

            decr_ac = decr_c.copy()
            decr_ac[0, cart_ind] -= 1

            decr_bc = decr_c.copy()
            decr_bc[1, cart_ind] -= 1

            PC_dir = (P - C)[cart_ind]
            la = (ia, ja, ka)[cart_ind]
            lb = (ib, jb, kb)[cart_ind]
            lc = (ic, jc, kc)[cart_ind] - 1

            decr_c = decr_c.flatten()
            decr_cc = decr_cc.flatten()
            decr_ac = decr_ac.flatten()
            decr_bc = decr_bc.flatten()

            return (
                p / (p + c) * PC_dir * recur(N + 1, *decr_c)
                + lc
                / (2 * c)
                * (recur(N, *decr_cc) - p / (p + c) * recur(N + 1, *decr_cc))
                + la / (2 * (p + c)) * recur(N + 1, *decr_ac)
                + lb / (2 * (p + c)) * recur(N + 1, *decr_bc)
            )

        def recur_vrr_aux_sph(cart_ind):
            assert (ib, jb, kb) == (0, 0, 0)
            decr_c = ang_moms2d.copy()
            decr_c[2, cart_ind] -= 1
            decr_ac = decr_c.copy()
            decr_ac[0, cart_ind] -= 1
            La = (ia, ja, ka)[cart_ind]
            PC_dir = (P - C)[cart_ind]
            return (
                rho
                / c
                * (
                    PC_dir * recur(N + 1, *decr_c.flatten())
                    + La / (2 * p) * recur(N + 1, *decr_ac.flatten())
                )
            )

        recur_vrr_aux_funcs = {
            "cart": recur_vrr_aux,
            "sph": recur_vrr_aux_sph,
        }
        recur_vrr_aux_func = recur_vrr_aux_funcs[cls.aux_vrr]

        # Base case
        if (ang_moms == 0).all():
            X_AB = A - B
            r2_PC = X_PC.dot(X_PC)
            r2_AB = X_AB.dot(X_AB)
            chi = rho * r2_PC
            K = exp(-mu * r2_AB)
            return 2 * pi**2.5 / sqrt(p + c) / (p * c) * K * boys(N, chi)
        elif ib > 0:
            return recur_hrr(0)
        elif jb > 0:
            return recur_hrr(1)
        elif kb > 0:
            return recur_hrr(2)
        elif ic > 0:
            return recur_vrr_aux_func(0)
        elif jc > 0:
            return recur_vrr_aux_func(1)
        elif kc > 0:
            return recur_vrr_aux_func(2)
        elif ia > 0:
            return recur_vrr(0)
        elif ja > 0:
            return recur_vrr(1)
        elif ka > 0:
            return recur_vrr(2)


class ThreeCenterTwoElectron(ThreeCenterTwoElectronBase):
    aux_vrr = "cart"


class ThreeCenterTwoElectronSph(ThreeCenterTwoElectronBase):
    aux_vrr = "sph"


class ThreeCenterTwoElectronShell(Function):
    @classmethod
    def eval(cls, La_tot, Lb_tot, Lc_tot, a, b, c, A, B, C):
        exprs = [
            ThreeCenterTwoElectron(*La, *Lb, *Lc, 0, a, b, c, A, B, C)
            for La, Lb, Lc in shell_iter((La_tot, Lb_tot, Lc_tot))
        ]
        # print(ThreeCenterTwoElectron.eval.cache_info())
        return exprs


class ThreeCenterTwoElectronSphShell(Function):
    @classmethod
    def eval(cls, La_tot, Lb_tot, Lc_tot, a, b, c, A, B, C):
        exprs = [
            ThreeCenterTwoElectronSph(*La, *Lb, *Lc, 0, a, b, c, A, B, C)
            for La, Lb, Lc in shell_iter((La_tot, Lb_tot, Lc_tot))
        ]
        # print(ThreeCenterTwoElectron.eval.cache_info())
        return exprs


class ERI(Function):
    """Variables named according to libreta paper [3]."""

    @classmethod
    def eval(
        cls, ia, ja, ka, ib, jb, kb, ic, jc, kc, id_, jd, kd, N, a, b, c, d, A, B, C, D
    ):
        ang_moms = np.array(
            (ia, ja, ka, ib, jb, kb, ic, jc, kc, id_, jd, kd), dtype=int
        )
        ang_moms2d = ang_moms.reshape(-1, 3)

        if any([am < 0 for am in ang_moms]):
            return 0

        AB = A - B
        xi = a + b
        P = (a * A + b * B) / xi  # Eq. (5)

        CD = C - D
        zeta = c + d
        Q = (c * C + d * D) / zeta  # Eq. (6)

        theta = xi * zeta / (xi + zeta)  # Eq. (7)

        def recur(N, *ang_moms):
            """Simple wrapper to pass all required arguments."""
            return ERI(*ang_moms, N, a, b, c, d, A, B, C, D)

        def recur_hrr(bra_or_ket, cart_ind):
            """Horizontal recursion relation to transfer angmoms in bra or ket.

            (a, b+1_x|cd) = (a + 1_x,b|cd) + X_AB (ab|cd)
            (ab|c, d + 1_x) = (ab|c + 1_x, d) + X_CD (ab|cd)
            """
            if bra_or_ket == "bra":
                XYZ = AB
                incr_ind = 0
            else:
                XYZ = CD
                incr_ind = 2

            decr_ind = incr_ind + 1
            incr_ang_moms = ang_moms2d.copy()
            incr_ang_moms[incr_ind, cart_ind] += 1  # Increase in left orbital
            incr_ang_moms[decr_ind, cart_ind] -= 1  # Decrease in right orbital
            incr_ang_moms = incr_ang_moms.flatten()
            decr_ang_moms = ang_moms2d.copy()
            decr_ang_moms[decr_ind, cart_ind] -= 1
            decr_ang_moms = decr_ang_moms.flatten()

            return recur(N, *incr_ang_moms) + XYZ[cart_ind] * recur(N, *decr_ang_moms)

        def recur_vrr(bra_or_ket, cart_ind):
            assert (ib, jb, kb) == (0, 0, 0)
            assert (id_, jd, kd) == (0, 0, 0)

            if bra_or_ket == "bra":
                XYZ = P - A
                decr_ind = 0
                decr_also_ind = 2
                exp1, exp2 = zeta, xi
                am1 = (ia, ja, ka)[cart_ind] - 1
                am2 = (ic, jc, kc)[cart_ind]
                sign = -1
            else:
                XYZ = Q - C
                decr_ind = 2
                decr_also_ind = 0
                exp1, exp2 = xi, zeta
                am1 = (ic, jc, kc)[cart_ind] - 1
                am2 = (ia, ja, ka)[cart_ind]
                sign = 1

            decr_ang_moms = ang_moms2d.copy()
            decr_ang_moms[decr_ind, cart_ind] -= 1
            decr_ang_moms = decr_ang_moms.flatten()
            decr2_ang_moms = ang_moms2d.copy()
            decr2_ang_moms[decr_ind, cart_ind] -= 2  # Further decrease angular momentum
            decr2_ang_moms = decr2_ang_moms.flatten()
            decr_also_ang_moms = ang_moms2d.copy()
            decr_also_ang_moms[decr_also_ind, cart_ind] -= 1
            decr_also_ang_moms = decr_also_ang_moms.flatten()

            denom = xi + zeta
            quot = exp1 / denom
            PQ = P - Q

            return (
                XYZ[cart_ind] * recur(N, *decr_ang_moms)
                + sign * quot * PQ[cart_ind] * recur(N + 1, *decr_ang_moms)
                + am1
                / (2 * exp2)
                * (recur(N, *decr2_ang_moms) - quot * recur(N + 1, *decr2_ang_moms))
                + am2 / (2 * denom) * recur(N + 1, *decr_also_ang_moms)
            )

        # Base case
        if all([am == 0 for am in ang_moms]):
            RAB2 = AB.dot(AB)
            RCD2 = CD.dot(CD)
            Kab = exp(-(a * b) / xi * RAB2)
            Kcd = exp(-(c * d) / zeta * RCD2)

            PQ = P - Q
            RPQ2 = PQ.dot(PQ)

            return (
                2
                * pi ** (5 / 2)
                / (xi * zeta + sqrt(xi + zeta))
                * Kab
                * Kcd
                * boys(N, theta * RPQ2)
            )  # Eq. (8) [ss|ss]^N
        #
        # Horizontal recursion relation to reduce angular momentum.
        #
        # Bra, basis func. with exp. b, centered at B with Lb_tot = (ib + jb + kb)
        elif ib > 0:
            return recur_hrr("bra", 0)
        elif jb > 0:
            return recur_hrr("bra", 1)
        elif kb > 0:
            return recur_hrr("bra", 2)
        # Ket, basis func. with exp. d, centered at D with Ld_tot = (id_ + jd + kd)
        elif id_ > 0:
            return recur_hrr("ket", 0)
        elif jd > 0:
            return recur_hrr("ket", 1)
        elif kd > 0:
            return recur_hrr("ket", 2)
        #
        # Vertical recursion relation to reduce angular momentum.
        #
        # Bra, basis func. with exp. a, centered at A with La_tot = (ia + ja + ka)
        elif ia > 0:
            return recur_vrr("bra", 0)
        elif ja > 0:
            return recur_vrr("bra", 1)
        elif ka > 0:
            return recur_vrr("bra", 2)
        # Ket, basis func. with exp. c, centered at C with Lc_tot = (ic + jc + kc)
        elif ic > 0:
            return recur_vrr("ket", 0)
        elif jc > 0:
            return recur_vrr("ket", 1)
        elif kc > 0:
            return recur_vrr("ket", 1)


class ERIShell(Function):
    @classmethod
    def eval(cls, La_tot, Lb_tot, Lc_tot, Ld_tot, a, b, c, d, A, B, C, D):
        exprs = [
            ERI(*La, *Lb, *Lc, *Ld, 0, a, b, c, d, A, B, C, D)
            for La, Lb, Lc, Ld in shell_iter((La_tot, Lb_tot, Lc_tot, Ld_tot))
        ]
        # print(ERI.eval.cache_info())
        return exprs


class SpinOrbit(Function):
    """1-electron spin-orbit interaction integral for Cartesian direction μ."""

    @classmethod
    @functools.cache
    def eval(cls, i, k, m, j, l, n, N, mu, a, b, A, B, C):
        ang_moms = (i, k, m, j, l, n)
        if any([am < 0 for am in ang_moms]):
            return 0

        p = a + b
        P = (a * A + b * B) / p
        X_PC = P - C

        def coulomb(N, *inds):
            return Coulomb(*inds, N, a, b, A, B, C)

        def recur(N, *inds):
            """Simple wrapper to pass all required arguments."""
            return SpinOrbit(*inds, N, mu, a, b, A, B, C)

        def decr(to_decr, decr_ind):
            one = np.zeros(3, dtype=int)
            one[decr_ind] = 1

            def Ni(inds):
                return inds[decr_ind]

            bra = np.array((i, k, m), dtype=int)
            ket = np.array((j, l, n), dtype=int)

            if to_decr == "bra":
                X = P - A
                bra[decr_ind] -= 1
            else:
                X = P - B
                ket[decr_ind] -= 1

            bra_decr = bra - one
            ket_decr = ket - one

            expr = (
                # The initial part of this recursion is the same as for the 1-electron
                # Coulomb integrals.
                X[decr_ind] * recur(N, *bra, *ket)
                - X_PC[decr_ind] * recur(N + 1, *bra, *ket)
                + 1
                / (2 * p)
                * Ni(bra)
                * (recur(N, *bra_decr, *ket) - recur(N + 1, *bra_decr, *ket))
                + 1
                / (2 * p)
                * Ni(ket)
                * (recur(N, *bra, *ket_decr) - recur(N + 1, *bra, *ket_decr))
                # From here on, the recursion differs.
            )

            if to_decr == "bra":
                X = B - C
                exponent = b
            else:
                X = A - C
                exponent = a
            # Second to line in Eq. (A34) in [1]
            expr += 2 * exponent * Matrix(one).cross(X)[mu] * coulomb(N + 1, *bra, *ket)
            # Last line in Eq. (A34) in [1]
            for k_ in range(
                3
            ):  # Use k_ instead of k, as this is already an angular momentum
                one_k = np.zeros(3, dtype=int)
                one_k[k_] = 1
                one_ind = np.cross(one, one_k)[mu]
                if to_decr == "bra":
                    expr += ket[k_] * one_ind * coulomb(N + 1, *bra, *(ket - one_k))
                else:
                    expr += bra[k_] * one_ind * coulomb(N + 1, *(bra - one_k), *ket)
            return expr

        # Base case
        if all([am == 0 for am in ang_moms]):
            AC = A - C
            BC = B - C
            ACxBC = AC.cross(BC)
            return 4 * p * ACxBC[m] * coulomb(N + 1, 0, 0, 0, 0, 0, 0)
        elif i > 0:
            return decr("bra", 0)
        elif j > 0:
            return decr("ket", 0)
        elif k > 0:
            return decr("bra", 1)
        elif l > 0:
            return decr("ket", 1)
        elif m > 0:
            return decr("bra", 2)
        elif n > 0:
            return decr("ket", 2)


class SpinOrbitShell(Function):
    @classmethod
    def eval(cls, La_tot, Lb_tot, a, b, A, B, C=(0.0, 0.0, 0.0)):
        exprs = [
            SpinOrbit(*La, *Lb, 0, mu, a, b, A, B, C)
            for (La, Lb), mu in it.product(shell_iter((La_tot, Lb_tot)), range(3))
        ]
        # print(SpinOrbit.eval.cache_info())
        return exprs


def get_center(i):
    symbs = [Symbol(str(i) + ind, real=True) for ind in ("x", "y", "z")]
    return Matrix([*symbs]).T  # Return column vector


def get_centers(i):
    symbs = [Symbol(str(i) + ind, real=True) for ind in ("x", "y", "z")]
    return Matrix([*symbs]).T  # Return column vector


def get_map(i, center_i):
    array = IndexedBase(i, shape=3)
    array_map = dict(zip(center_i, array))
    return array, array_map


def cart2spherical(L_tots, exprs):
    assert len(L_tots) > 0

    # Coefficient matrices for Cartesian-to-spherical conversion
    coeffs = [Array(CART2SPH[L]) for L in L_tots]
    cart_shape = [(l + 1) * (l + 2) // 2 for l in L_tots]
    cart = Array(exprs).reshape(*cart_shape)

    sph = tc(tp(coeffs[0], cart), (1, 2))
    if len(L_tots) == 2:
        sph = tc(tp(sph, coeffs[1].transpose()), (1, 2))
    elif len(L_tots) == 3:
        _, Cb, Cc = coeffs
        sph = tc(tp(permutedims(sph, (0, 2, 1)), Cb.transpose()), (2, 3))
        sph = tc(tp(permutedims(sph, (0, 2, 1)), Cc.transpose()), (2, 3))
    else:
        raise Exception(
            "Cartesian -> spherical transformation for 4-center integrals "
            "is not implemented!"
        )

    # Cartesian-to-spherical transformation introduces quite a number of
    # multiplications by 1.0, which are uneccessary. Here, we try to drop
    # some of them by replacing numbers very close to +1.0 with 1.
    sph = sph.replace(lambda n: n.is_Number and (abs(n - 1) <= ONE_THRESH), lambda n: 1)
    # TODO: maybe something along the lines
    # sph = map(lambda expr: expr.evalf(), flatten(sph))
    # is faster?
    return flatten(sph)


def gen_integral_exprs(
    int_func,
    L_maxs,
    kind,
    maps=None,
    sph=False,
):
    if maps is None:
        maps = list()

    ranges = [range(L + 1) for L in L_maxs]

    for L_tots in it.product(*ranges):
        time_str = time.strftime("%H:%M:%S")
        start = datetime.now()
        print(f"{time_str} - Generating {L_tots} {kind}")
        sys.stdout.flush()
        # Generate actual list of expressions.
        exprs = int_func(*L_tots)
        print("\t... generated expressions")
        sys.stdout.flush()
        # if sph:
        # exprs = cart2spherical(L_tots, exprs)
        # print("\t... did Cartesian -> Spherical conversion")
        # sys.stdout.flush()

        # Common subexpression elimination
        repls, reduced = cse(list(exprs), order="none")
        print("\t... did common subexpression elimination")

        # Replacement expressions, used to form the reduced expressions.
        for i, (lhs, rhs) in enumerate(repls):
            rhs = rhs.evalf()
            # Replace occurences of Ax, Ay, Az, ... with A[0], A[1], A[2], ...
            rhs = functools.reduce(lambda rhs, map_: rhs.xreplace(map_), maps, rhs)
            repls[i] = (lhs, rhs)

        # Reduced expression, i.e., the final integrals/expressions.
        for i, red in enumerate(reduced):
            red = red.evalf()
            reduced[i] = functools.reduce(
                lambda red, map_: red.xreplace(map_), maps, red
            )
        # Carry out Cartesian-to-spherical transformation, if requested.
        if sph:
            reduced = cart2spherical(L_tots, reduced)
            print("\t... did Cartesian -> Spherical conversion")
            sys.stdout.flush()

        dur = datetime.now() - start
        print(f"\t... finished in {str(dur)} h")
        sys.stdout.flush()
        yield (repls, reduced), L_tots


def render_py_funcs(exprs_Ls, args, base_name, doc_func, add_imports=None, comment=""):
    if add_imports is None:
        add_imports = ()
    add_imports_str = "\n".join(add_imports)
    if add_imports:
        add_imports_str += "\n\n"

    args = ", ".join((map(str, args)))

    py_funcs = list()
    for (repls, reduced), L_tots in exprs_Ls:
        doc_str = doc_func(L_tots)
        doc_str += "\n\nGenerated code; DO NOT modify by hand!"
        name = base_name + "_" + "".join(str(l) for l in L_tots)
        print(f"Rendering '{name}' ... ", end="")
        start = time.time()
        py_func = make_py_func(repls, reduced, args=args, name=name, doc_str=doc_str)
        dur = time.time() - start
        print(f"finished in {dur: >8.2f} s")
        py_funcs.append(py_func)
    py_funcs_joined = "\n\n".join(py_funcs)

    if comment != "":
        comment = f'"""\n{comment}\n"""\n\n'

    np_tpl = Template(
        "import numpy\n\n{{ add_imports }}{{ comment }}{{ py_funcs }}",
        trim_blocks=True,
        lstrip_blocks=True,
    )
    np_rendered = np_tpl.render(
        comment=comment, add_imports=add_imports_str, py_funcs=py_funcs_joined
    )
    np_rendered = textwrap.dedent(np_rendered)
    try:
        from black import format_str, FileMode

        np_rendered = format_str(np_rendered, mode=FileMode(line_length=90))
    except ModuleNotFoundError:
        print("Skipped formatting with black, as it is not installed!")

    return np_rendered


def render_c_funcs(exprs_Ls, args, base_name, doc_func, add_imports=None, comment=""):
    if add_imports is not None:
        raise Exception("Implement me!")

    arg_strs = [str(arg) for arg in args]

    funcs = list()
    signatures = list()
    for (repls, reduced), L_tots in exprs_Ls:
        doc_str = doc_func(L_tots)
        doc_str += "\n\n\t\tGenerated code; DO NOT modify by hand!"
        doc_str = textwrap.dedent(doc_str)
        name = base_name + "_" + "".join(str(l) for l in L_tots)
        func, signature = make_c_func(
            repls, reduced, args=arg_strs, name=name, doc_str=doc_str
        )
        funcs.append(func)
        signatures.append(signature)
    funcs_joined = "\n\n".join(funcs)

    if comment != "":
        comment = f"/*{comment}*/"

    # Render C files
    c_tpl = Template(
        "#include <math.h>\n\n{{ comment }}\n\n{{ funcs }}",
        trim_blocks=True,
        lstrip_blocks=True,
    )
    c_rendered = c_tpl.render(comment=comment, funcs=funcs_joined)
    c_rendered = textwrap.dedent(c_rendered)
    # Render simple header file
    h_rendered = "\n".join([f"{sig};" for sig in signatures])
    return c_rendered, h_rendered


def write_file(out_dir, name, rendered):
    out_name = out_dir / name
    with open(out_name, "w") as handle:
        handle.write(rendered)
    print(f"Wrote '{out_name}'.")


def write_render(
    ints_Ls,
    args,
    name,
    doc_func,
    out_dir,
    comment="",
    py_kwargs=None,
    c=True,
    c_kwargs=None,
):
    if py_kwargs is None:
        py_kwargs = {}
    if c_kwargs is None:
        c_kwargs = {}
    ints_Ls = list(ints_Ls)
    # Python
    py_rendered = render_py_funcs(
        ints_Ls, args, name, doc_func, comment=comment, **py_kwargs
    )
    write_file(out_dir, f"{name}.py", py_rendered)
    # C
    if c:
        c_rendered, h_rendered = render_c_funcs(
            ints_Ls,
            args,
            name,
            doc_func,
            comment=comment,
            **c_kwargs,
        )
        write_file(out_dir, f"{name}.c", c_rendered)
        write_file(out_dir, f"{name}.h", h_rendered)


def parse_args(args):
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--lmax",
        default=L_MAX,
        type=int,
        help="Generate 1e-integrals up to this maximum angular momentum.",
    )
    parser.add_argument(
        "--lauxmax",
        default=L_AUX_MAX,
        type=int,
        help="Maximum angular moment for integrals using auxiliary functions.",
    )
    parser.add_argument(
        "--write",
        action="store_true",
        help="Write out generated integrals to the current directory, potentially overwriting the present modules.",
    )
    parser.add_argument(
        "--out-dir",
        default="devel_ints",
        help="Directory, where the generated integrals are written.",
    )
    keys_str = f"({', '.join(KEYS)})"
    parser.add_argument(
        "--keys",
        nargs="+",
        help=f"Generate only certain expressions. Possible keys are: {keys_str}. "
        "If not given, all expressions are generated.",
    )

    return parser.parse_args()


def run():
    args = parse_args(sys.argv[1:])

    l_max = args.lmax
    l_aux_max = args.lauxmax
    out_dir = Path(args.out_dir if not args.write else ".")
    keys = args.keys
    if keys is None:
        keys = list()
    try:
        os.mkdir(out_dir)
    except FileExistsError:
        pass

    try:
        global CART2SPH
        CART2SPH = cart2sph_coeffs(max(l_max, l_aux_max), zero_small=True)
    except NameError:
        print("cart2sph_coeffs import is deactivated or pysisyphus is not installed.")

    # Cartesian basis function centers A and B.
    center_A = get_center("A")
    center_B = get_center("B")
    center_C = get_center("C")
    center_D = get_center("D")
    # center_R = get_center("R")
    Xa, Ya, Za = symbols("Xa Ya Za")

    # Orbital exponents a, b, c, d.
    a, b, c, d = symbols("a b c d", real=True)

    # These maps will be used to convert {Ax, Ay, ...} to array quantities
    # in the generated code. This way an iterable/np.ndarray can be used as
    # function argument instead of (Ax, Ay, Az, Bx, By, Bz).
    A, A_map = get_map("A", center_A)
    B, B_map = get_map("B", center_B)
    C, C_map = get_map("C", center_C)
    D, D_map = get_map("D", center_D)

    boys_import = ("from pysisyphus.wavefunction.ints.boys import boys",)

    #################
    # Cartesian GTO #
    #################

    def cart_gto():
        def cart_gto_doc_func(L_tot):
            (La_tot,) = L_tot
            shell_a = L_MAP[La_tot]
            return (
                f"3D Cartesian {shell_a}-Gaussian shell.\n\n"
                "Exponent a, centered at A, evaluated at (Xa, Ya, Za) + A."
            )

        # This code can evaluate multiple points at a time
        cart_gto_Ls = gen_integral_exprs(
            lambda La_tot: CartGTOShell(La_tot, a, Xa, Ya, Za),
            (l_max,),
            "cart_gto",
        )
        cart_gto_rendered = render_py_funcs(
            cart_gto_Ls, (a, Xa, Ya, Za), "cart_gto3d", cart_gto_doc_func
        )
        write_file(out_dir, "gto3d.py", cart_gto_rendered)
        print()

    #####################
    # Overlap integrals #
    #####################

    def overlap():
        def ovlp_doc_func(L_tots):
            La_tot, Lb_tot = L_tots
            shell_a = L_MAP[La_tot]
            shell_b = L_MAP[Lb_tot]
            return f"Cartesian 3D ({shell_a}{shell_b}) overlap integral."

        ovlp_ints_Ls = gen_integral_exprs(
            lambda La_tot, Lb_tot: Overlap3dShell(
                La_tot, Lb_tot, a, b, center_A, center_B
            ),
            (l_max, l_max),
            "overlap",
            (A_map, B_map),
        )
        write_render(
            ovlp_ints_Ls, (a, A, b, B), "ovlp3d", ovlp_doc_func, out_dir, c=True
        )
        print()

    ###########################
    # Dipole moment integrals #
    ###########################

    def dipole():
        def dipole_doc_func(L_tots):
            La_tot, Lb_tot = L_tots
            shell_a = L_MAP[La_tot]
            shell_b = L_MAP[Lb_tot]
            return (
                f"Cartesian 3D ({shell_a}{shell_b}) dipole moment integrals.\n"
                "The origin is at C."
            )

        dipole_comment = """
        Dipole integrals are given in the order:
        for bf_a in basis_functions_a:
            for bf_b in basis_functions_b:
                for cart_dir in (x, y, z):
                    dipole_integrals(bf_a, bf_b, cart_dir)

        So for <s_a|μ|s_b> it will be:

            <s_a|x|s_b>
            <s_a|y|s_b>
            <s_a|z|s_b>
        """

        dipole_ints_Ls = gen_integral_exprs(
            lambda La_tot, Lb_tot: Multipole3dShell(
                La_tot, Lb_tot, a, b, center_A, center_B, 1, center_C
            ),
            (l_max, l_max),
            "dipole moment",
            (A_map, B_map, C_map),
        )
        write_render(
            dipole_ints_Ls,
            (a, A, b, B, C),
            "dipole3d",
            dipole_doc_func,
            out_dir,
            comment=dipole_comment,
            c=True,
        )
        print()

    ###########################################
    # Diagonal of quadrupole moment integrals #
    ###########################################

    def diag_quadrupole():
        def diag_quadrupole_doc_func(L_tots):
            La_tot, Lb_tot = L_tots
            shell_a = L_MAP[La_tot]
            shell_b = L_MAP[Lb_tot]
            return (
                f"Cartesian 3D ({shell_a}{shell_b}) quadrupole moment integrals\n"
                "for operators x², y² and z². The origin is at C."
            )

        diag_quadrupole_comment = """
        Diagonal of the quadrupole moment matrix with operators x², y², z².

        for rr in (xx, yy, zz):
            for bf_a in basis_functions_a:
                for bf_b in basis_functions_b:
                        quadrupole_integrals(bf_a, bf_b, rr)
        """

        diag_quadrupole_ints_Ls = gen_integral_exprs(
            lambda La_tot, Lb_tot: DiagQuadrupole3dShell(
                La_tot, Lb_tot, a, b, center_A, center_B, center_C
            ),
            (l_max, l_max),
            "diag quadrupole moment",
            (A_map, B_map, C_map),
        )
        write_render(
            diag_quadrupole_ints_Ls,
            (a, A, b, B, C),
            "diag_quadrupole3d",
            diag_quadrupole_doc_func,
            out_dir,
            comment=diag_quadrupole_comment,
            c=True,
        )
        print()

    ###############################
    # Quadrupole moment integrals #
    ###############################

    def quadrupole():
        def quadrupole_doc_func(L_tots):
            La_tot, Lb_tot = L_tots
            shell_a = L_MAP[La_tot]
            shell_b = L_MAP[Lb_tot]
            return (
                f"Cartesian 3D ({shell_a}{shell_b}) quadrupole moment integrals.\n"
                "The origin is at C."
            )

        quadrupole_comment = """
        Quadrupole integrals contain the upper triangular part of the symmetric
        3x3 quadrupole matrix.

        / xx xy xz \\
        |    yy yz |
        \       zz /
        """

        quadrupole_ints_Ls = gen_integral_exprs(
            lambda La_tot, Lb_tot: Multipole3dShell(
                La_tot, Lb_tot, a, b, center_A, center_B, 2, center_C
            ),
            (l_max, l_max),
            "quadrupole moment",
            (A_map, B_map, C_map),
        )

        write_render(
            quadrupole_ints_Ls,
            (a, A, b, B, C),
            "quadrupole3d",
            quadrupole_doc_func,
            out_dir,
            comment=quadrupole_comment,
            c=True,
        )
        print()

    ############################
    # Kinetic energy integrals #
    ############################

    def kinetic():
        def kinetic_doc_func(L_tots):
            La_tot, Lb_tot = L_tots
            shell_a = L_MAP[La_tot]
            shell_b = L_MAP[Lb_tot]
            return f"Cartesian 3D ({shell_a}{shell_b}) kinetic energy integral."

        kinetic_ints_Ls = gen_integral_exprs(
            lambda La_tot, Lb_tot: Kinetic3dShell(
                La_tot, Lb_tot, a, b, center_A, center_B
            ),
            (l_max, l_max),
            "kinetic",
            (A_map, B_map),
        )
        write_render(
            kinetic_ints_Ls,
            (a, A, b, B),
            "kinetic3d",
            kinetic_doc_func,
            out_dir,
            c=True,
        )
        print()

    #########################
    # 1el Coulomb Integrals #
    #########################

    def coulomb():
        def coulomb_doc_func(L_tots):
            La_tot, Lb_tot = L_tots
            shell_a = L_MAP[La_tot]
            shell_b = L_MAP[Lb_tot]
            return f"Cartesian ({shell_a}{shell_b}) 1-electron Coulomb integral."

        coulomb_ints_Ls = gen_integral_exprs(
            lambda La_tot, Lb_tot: CoulombShell(
                La_tot, Lb_tot, a, b, center_A, center_B, center_C
            ),
            (l_max, l_max),
            "coulomb3d",
            (A_map, B_map, C_map),
        )
        coulomb_rendered = render_py_funcs(
            coulomb_ints_Ls,
            (a, A, b, B, C),
            "coulomb3d",
            coulomb_doc_func,
            add_imports=boys_import,
        )
        # TODO: handle add_args and Boys function in C
        write_file(out_dir, "coulomb3d.py", coulomb_rendered)
        print()

    ###############################################
    # Two-center two-electron repulsion integrals #
    ###############################################

    def _2center2electron():
        def _2center2el_doc_func(L_tots):
            La_tot, Lb_tot = L_tots
            shell_a = L_MAP[La_tot]
            shell_b = L_MAP[Lb_tot]
            return (
                f"Cartesian ({shell_a}|{shell_b}) "
                "two-center two-electron repulsion integral."
            )

        _2center2el_ints_Ls = gen_integral_exprs(
            lambda La_tot, Lb_tot: TwoCenterTwoElectronShell(
                La_tot,
                Lb_tot,
                a,
                b,
                center_A,
                center_B,
            ),
            (l_aux_max, l_aux_max),
            "_2center2el3d",
            (A_map, B_map),
        )
        write_render(
            _2center2el_ints_Ls,
            (a, A, b, B),
            "_2center2el3d",
            _2center2el_doc_func,
            out_dir,
            c=False,
            py_kwargs={"add_imports": boys_import},
        )
        print()

    #################################################
    # Three-center two-electron repulsion integrals #
    #################################################

    def _3center2electron():
        def _3center2el_doc_func(L_tots):
            La_tot, Lb_tot, Lc_tot = L_tots
            shell_a = L_MAP[La_tot]
            shell_b = L_MAP[Lb_tot]
            shell_c = L_MAP[Lc_tot]
            return (
                f"Cartesian ({shell_a}{shell_b}|{shell_c}) "
                "three-center two-electron repulsion integral."
            )

        _3center2el_ints_Ls = gen_integral_exprs(
            lambda La_tot, Lb_tot, Lc_tot: ThreeCenterTwoElectronShell(
                La_tot, Lb_tot, Lc_tot, a, b, c, center_A, center_B, center_C
            ),
            (l_max, l_max, l_aux_max),
            "_3center2el3d",
            (A_map, B_map, C_map),
        )
        write_render(
            _3center2el_ints_Ls,
            (a, A, b, B, c, C),
            "_3center2el3d",
            _3center2el_doc_func,
            out_dir,
            c=False,
            py_kwargs={"add_imports": boys_import},
        )
        print()

    def _3center2electron_sph():
        def _3center2el_doc_func(L_tots):
            La_tot, Lb_tot, Lc_tot = L_tots
            shell_a = L_MAP[La_tot]
            shell_b = L_MAP[Lb_tot]
            shell_c = L_MAP[Lc_tot]
            return (
                f"Cartesian ({shell_a}{shell_b}|{shell_c}) "
                "three-center two-electron repulsion integral for conversion to "
                "spherical harmonics.\nUses Ahlrichs' vertical recursion relation, "
                "that leaves out some terms, that cancel\nwhen convertig to "
                "spherical harmonics."
            )

        _3center2el_ints_Ls = gen_integral_exprs(
            lambda La_tot, Lb_tot, Lc_tot: ThreeCenterTwoElectronSphShell(
                La_tot, Lb_tot, Lc_tot, a, b, c, center_A, center_B, center_C
            ),
            (l_max, l_max, l_aux_max),
            "_3center2el3d_sph",
            (A_map, B_map, C_map),
            sph=False,
        )
        write_render(
            _3center2el_ints_Ls,
            (a, A, b, B, c, C),
            "_3center2el3d_sph",
            _3center2el_doc_func,
            out_dir,
            c=False,
            py_kwargs={"add_imports": boys_import},
        )
        print()

    ####################################
    # Spin-orbit interaction integrals #
    ####################################

    # def so1el_doc_func(L_tots):
    # La_tot, Lb_tot = L_tots
    # shell_a = L_MAP[La_tot]
    # shell_b = L_MAP[Lb_tot]
    # return f"Cartesian ({shell_a}{shell_b}) 1-electron spin-orbit-interaction integral."

    # I think this is still faulty!
    # so1el_ints_Ls = gen_integral_exprs(
    # lambda La_tot, Lb_tot: SpinOrbitShell(
    # La_tot, Lb_tot, a, b, center_A, center_B, center_C
    # ),
    # (l_max, l_max),
    # "so1el",
    # (A_map, B_map, C_map),
    # )

    # so1el_rendered = render_py_funcs(
    # so1el_ints_Ls,
    # (a, A, b, B, C),
    # "so1el",
    # so1el_doc_func,
    # add_imports=boys_import,
    # )
    # write_file(out_dir, "so1el.py", so1el_rendered)
    # print()

    #########################
    # 2el Coulomb Integrals #
    #########################

    # def eri_doc_func(L_tots):
    # La_tot, Lb_tot, Lc_tot, Ld_tot = L_tots
    # shell_a = L_MAP[La_tot]
    # shell_b = L_MAP[Lb_tot]
    # shell_c = L_MAP[Lc_tot]
    # shell_d = L_MAP[Ld_tot]
    # return (
    # f"Cartesian [{shell_a}{shell_b}|{shell_c}{shell_d}] "
    # "2-electron electron repulsion integral."
    # )

    # I think this is still faulty!
    # eri_ints_Ls = gen_integral_exprs(
    # lambda La_tot, Lb_tot, Lc_tot, Ld_tot: ERIShell(
    # La_tot,
    # Lb_tot,
    # Lc_tot,
    # Ld_tot,
    # a,
    # b,
    # c,
    # d,
    # center_A,
    # center_B,
    # center_C,
    # center_D,
    # ),
    # # (l_max, l_max, l_max, l_max),
    # (2, 2, 2, 2),  # Stop at [dd|dd] for now
    # "eri",
    # (A_map, B_map, C_map, D_map),
    # )
    # eri_rendered = render_py_funcs(
    # eri_ints_Ls,
    # (a, A, b, B, c, C, d, D),
    # "eri",
    # eri_doc_func,
    # add_imports=boys_import,
    # )
    # write_file(out_dir, "eri.py", eri_rendered)
    # print()

    funcs = {
        "cgto": cart_gto,  # Cartesian Gaussian-type-orbital for density evaluation
        "ovlp": overlap,  # Overlap integrals
        "dpm": dipole,  # Linear moment (dipole) integrals
        "dqpm": diag_quadrupole,  # Diagonal part of the quadrupole tensor
        "qpm": quadrupole,  # Quadratic moment (quadrupole) integrals
        "kin": kinetic,  # Kinetic energy integrals
        "coul": coulomb,  # 1-electron Coulomb integrals
        # Integrals for density fitting
        "2c2e": _2center2electron,  # 2-center-2-electron integrals for DF
        "3c2e": _3center2electron,  # 3-center-2-electron integrals for density fitting
        "3c2e_sph": _3center2electron_sph,  # 3c2el for spherical transformation for DF
    }
    if len(keys) == 0:
        keys = funcs.keys()
    for key in keys:
        funcs[key]()


if __name__ == "__main__":
    run()



from matplotlib import cm
import matplotlib.animation as animation
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import numpy as np
from sympy import symbols, diff, lambdify, sympify

from pysisyphus.calculators.Calculator import Calculator
from pysisyphus.Geometry import Geometry
from pysisyphus.interpolate import interpolate
from pysisyphus.plotters.AnimPlot import AnimPlot


class AnaPotBase(Calculator):
    def __init__(
        self,
        V_str,
        scale=1.0,
        xlim=(-1, 1),
        ylim=(-1, 1),
        levels=None,
        use_sympify=True,
        minima=None,
        saddles=None,
    ):
        super().__init__()
        self.V_str = V_str
        self.scale = scale
        self.xlim = xlim
        self.ylim = ylim
        if levels is not None:
            levels = levels * self.scale
        self.levels = levels
        if minima is None:
            minima = list()
        self.minima = np.array(minima, dtype=float)
        if saddles is None:
            saddles = list()
        self.saddles = np.array(saddles, dtype=float)

        x, y = symbols("x y")
        if use_sympify:
            V = sympify(V_str)
        else:
            V = V_str
        dVdx = diff(V, x)
        dVdy = diff(V, y)
        self.V = lambdify((x, y), V, "numpy")
        self.dVdx = lambdify((x, y), dVdx, "numpy")
        self.dVdy = lambdify((x, y), dVdy, "numpy")

        dVdxdx = diff(V, x, x)
        dVdxdy = diff(V, x, y)
        dVdydy = diff(V, y, y)

        self.dVdxdx = lambdify((x, y), dVdxdx, "numpy")
        self.dVdxdy = lambdify((x, y), dVdxdy, "numpy")
        self.dVdydy = lambdify((x, y), dVdydy, "numpy")

        self.fake_atoms = ("X",)  # X, dummy atom

        self.analytical_2d = True
        self.energy_calcs = 0
        self.forces_calcs = 0
        self.hessian_calcs = 0

        # Dummies
        self.mult = 1
        self.charge = 0

        self.fig = None
        self.ax = None

    def get_energy(self, atoms, coords):
        self.energy_calcs += 1
        x, y, z = coords
        energy = self.scale * self.V(x, y)
        return {
            "energy": energy,
        }

    def get_forces(self, atoms, coords):
        self.forces_calcs += 1
        x, y, z = coords
        dVdx = self.dVdx(x, y)
        dVdy = self.dVdy(x, y)
        dVdz = np.zeros_like(dVdx)
        forces = self.scale * -np.array((dVdx, dVdy, dVdz))
        results = {
            "forces": forces,
        }
        results.update(self.get_energy(atoms, coords))
        return results

    def get_hessian(self, atoms, coords):
        self.hessian_calcs += 1
        x, y, z = coords
        dVdxdx = self.dVdxdx(x, y)
        dVdxdy = self.dVdxdy(x, y)
        dVdydy = self.dVdydy(x, y)
        hessian = self.scale * np.array(
            ((dVdxdx, dVdxdy, 0), (dVdxdy, dVdydy, 0), (0, 0, 0))
        )
        results = {
            "hessian": hessian,
        }
        results.update(self.get_energy(atoms, coords))
        return results

    def statistics(self):
        return (
            f"Energy calculations: {self.energy_calcs}, Force calculations: "
            f"{self.forces_calcs}, Hessian calculations: {self.hessian_calcs}"
        )

    def plot(self, levels=None, show=False, **figkwargs):
        if not self.fig or not self.ax:
            self.fig, self.ax = plt.subplots(**figkwargs)

        x = np.linspace(*self.xlim, 100)
        y = np.linspace(*self.ylim, 100)
        X, Y = np.meshgrid(x, y)
        Z = np.full_like(X, 0)
        pot_coords = np.stack((X, Y, Z))
        pot = self.get_energy(self.fake_atoms, pot_coords)["energy"]

        if levels is None:
            if self.levels is None:
                levels = np.linspace(pot.min(), pot.max(), 35)
            else:
                levels = self.levels

        # Draw the contourlines of the potential
        contours = self.ax.contour(X, Y, pot, levels)

        # Avoid duplicated colorbars
        if len(self.fig.axes) == 1:
            self.fig.colorbar(contours)

        if show:
            plt.show()

    def plot3d(
        self,
        levels=None,
        show=False,
        zlim=None,
        vmin=None,
        vmax=None,
        resolution=100,
        rcount=50,
        ccount=50,
        nan_above=None,
        init_view=None,
        colorbar=False,
        **figkwargs,
    ):
        self.fig = plt.figure(**figkwargs)
        self.ax = self.fig.add_subplot(111, projection="3d")
        x = np.linspace(*self.xlim, resolution)
        y = np.linspace(*self.ylim, resolution)
        X, Y = np.meshgrid(x, y)
        Z = np.full_like(X, 0)
        pot_coords = np.stack((X, Y, Z))
        pot = self.get_energy(self.fake_atoms, pot_coords)["energy"]
        if nan_above:
            pot[pot > nan_above] = np.nan

        if vmin is None:
            vmin = np.nanmin(pot)
        if vmax is None:
            vmax = 0.125 * np.nanmax(pot)
        surf = self.ax.plot_surface(
            X,
            Y,
            pot,
            rcount=rcount,
            ccount=ccount,
            cmap=cm.coolwarm,
            vmin=vmin,
            vmax=vmax,
        )
        if zlim is not None:
            self.ax.set_zlim(*zlim)

        if colorbar:
            cb = self.fig.colorbar(surf, shrink=0.45, pad=0.0)
            cb.set_label("f(x,y)")

        if init_view:
            self.ax.view_init(*init_view)

        if show:
            plt.show()

        return X, Y, pot

    def plot_eigenvalue_structure(self, grid=50, levels=None, show=False):
        self.plot(levels=levels)
        xs = np.linspace(*self.xlim, grid)
        ys = np.linspace(*self.ylim, grid)
        X, Y = np.meshgrid(xs, ys)
        z = list()
        for x_, y_ in zip(X.flatten(), Y.flatten()):
            H = self.get_hessian(self.fake_atoms, (x_, y_, 0))["hessian"]
            w, v = np.linalg.eigh(H)
            z.append(1 if (w < 0).any() else 0)
        Z = np.array(z).reshape(X.shape)
        self.ax.contourf(X, Y, Z, cmap=cm.Reds)
        if show:
            plt.show()

    def plot_coords(self, xs, ys, enum=True, show=False, title=None):
        self.plot()
        self.ax.plot(xs, ys, "o-")
        if enum:
            for i, (x, y) in enumerate(zip(xs, ys)):
                self.ax.annotate(i, (x, y))
        if title:
            self.fig.suptitle(title)
        if show:
            plt.show()

    def plot_opt(self, opt, *args, **kwargs):
        xs, ys = np.array(opt.coords).T[:2]
        self.plot_coords(xs, ys, *args, **kwargs)

    def plot_geoms(self, geoms, **kwargs):
        coords = np.array([geom.coords for geom in geoms])
        xs, ys = coords.T[:2]
        self.plot_coords(xs, ys, **kwargs)

    def plot_irc(self, irc, *args, **kwargs):
        xs, ys = irc.all_coords.T[:2]
        self.plot_coords(xs, ys, *args, **kwargs)

    def anim_opt(
        self, opt, energy_profile=False, colorbar=False, figsize=(8, 6), show=False
    ):
        try:
            min_ = self.levels.min()
            max_ = self.levels.max()
            num = self.levels.size
            levels = (min_, max_, num)
        except TypeError:
            levels = None

        anim = AnimPlot(
            self.__class__(),
            opt,
            xlim=self.xlim,
            ylim=self.ylim,
            levels=levels,
            energy_profile=energy_profile,
            colorbar=colorbar,
            figsize=figsize,
        )
        anim.animate()
        if show:
            plt.show()
        return anim

    def anim_coords(self, coords, interval=50, show=False, title_func=None):
        self.plot()
        steps = range(len(coords))
        scatter = self.ax.scatter(*coords[0][:2], s=20)

        def func(frame):
            if title_func:
                self.ax.set_title(title_func(frame))
            scatter.set_offsets(coords[frame][:2])

        self.animation = animation.FuncAnimation(
            self.fig, func, frames=steps, interval=interval
        )

        if show:
            plt.show()

    @classmethod
    def get_geom(
        cls, coords, atoms=("X",), V_str=None, calc_kwargs=None, geom_kwargs=None
    ):
        if calc_kwargs is None:
            calc_kwargs = dict()
        if geom_kwargs is None:
            geom_kwargs = dict()

        geom = Geometry(atoms, coords, **geom_kwargs)
        if V_str:
            calc_kwargs["V_str"] = V_str
        geom.set_calculator(cls(**calc_kwargs))
        return geom

    def get_path(self, num, minima_inds=None):
        between = num - 2

        inds = 0, 1
        if minima_inds is not None:
            inds = minima_inds
        initial_ind, final_ind = inds

        initial_geom = self.get_geom(self.minima[initial_ind])
        final_geom = self.get_geom(self.minima[final_ind])
        geoms = interpolate(initial_geom, final_geom, between=between)
        for geom in geoms:
            # Creating new instances can be really slow when the sympy calls
            # need some time. For now we just reuse the current calculator...
            # geom.set_calculator(self.__class__())
            geom.set_calculator(self)
        return geoms

    def get_geoms_from_stored_coords(
        self, coords_list, i=None, calc_kwargs=None, geom_kwargs=None
    ):
        if calc_kwargs is None:
            calc_kwargs = dict()
        if geom_kwargs is None:
            geom_kwargs = dict()
        geoms = [
            self.get_geom(coords, calc_kwargs=calc_kwargs, geom_kwargs=geom_kwargs)
            for coords in coords_list
        ]
        if i is not None:
            return geoms[i]
        else:
            return geoms

    def get_minima(self, i=None, calc_kwargs=None, geom_kwargs=None):
        return self.get_geoms_from_stored_coords(
            self.minima, i=i, calc_kwargs=calc_kwargs, geom_kwargs=geom_kwargs
        )

    def get_saddles(self, i=None, calc_kwargs=None, geom_kwargs=None):
        return self.get_geoms_from_stored_coords(
            self.saddles, i=i, calc_kwargs=calc_kwargs, geom_kwargs=geom_kwargs
        )



import h5py
import numpy as np

from pysisyphus.drivers.opt import run_opt
from pysisyphus.helpers_pure import highlight_text
from pysisyphus.optimizers.RFOptimizer import RFOptimizer
from pysisyphus.TablePrinter import TablePrinter


def relaxed_scan(
    geom,
    calc_getter,
    constrain_prims,
    target_values,
    title,
    max_cycles=25,
    trust_radius=0.5,
    thresh=1e-2,
    dump=True,
):
    """Relaxed scan, allowing fixing of multiple primitive internals."""

    # Constrain desired primitives
    copy_kwargs = {
        "coord_type": "redund",
        "coord_kwargs": {"constrain_prims": constrain_prims},
    }
    scan_geom = geom.copy(**copy_kwargs)
    calc = calc_getter()
    scan_geom.set_calculator(calc)

    lengths_ = [len(inds) for prim_type, *inds in constrain_prims]
    if all([True for l in lengths_ if l == 2]):
        unit = "au"
    elif all([True for l in lengths_ if l in (3, 4)]):
        unit = "rad"
    else:
        unit = "au (rad)"

    typed_prims = scan_geom.internal.typed_prims
    constr_inds = [typed_prims.index(cp) for cp in constrain_prims]
    constr_fmt = lambda arr: f"({np.array2string(arr, precision=4)}) {unit}"

    def print_constraints():
        print(f"Desired constraints: {constr_fmt(target_values)}")
        print(f" Actual constraints: {constr_fmt(scan_geom.coords[constr_inds])}")

    trj = ""
    scan_cart_coords = [scan_geom.cart_coords.copy()]
    scan_energies = [scan_geom.energy]
    actual_values = [scan_geom.coords[constr_inds]]

    for i in range(max_cycles):
        print(highlight_text(f"Step {i}", level=1))
        cur_diff = target_values - scan_geom.coords[constr_inds]
        step = cur_diff
        step_norm = np.linalg.norm(step)
        print_constraints()
        print(f"         Difference: {step_norm:.6f} {unit}\n")
        if step_norm <= thresh:
            print(
                f"Relaxed scan converged! Norm of proposed step <= {thresh:.4f} {unit}"
            )
            break
        if step_norm > trust_radius:
            step = trust_radius * step / step_norm
        new_coords = scan_geom.coords.copy()
        new_coords[constr_inds] += step
        scan_geom.set_coords(new_coords, update_constraints=True)

        prefix = f"{title}_step_{i}"
        opt_kwargs = {
            "prefix": prefix,
            "h5_group_name": prefix,
            "dump": dump,
        }
        opt = RFOptimizer(scan_geom, **opt_kwargs)
        opt.run()

        scan_cart_coords.append(scan_geom.cart_coords.copy())
        scan_energies.append(scan_geom.energy)
        actual_values.append(scan_geom.coords[constr_inds])
        trj += scan_geom.as_xyz() + "\n"

    scan_cart_coords = np.array(scan_cart_coords)
    scan_energies = np.array(scan_energies)
    actual_values = np.array(actual_values)

    if dump:
        trj_fn = f"{title}_scan_trj.xyz"
        with open(trj_fn, "w") as handle:
            handle.write(trj)
        print(f"Dumped optimized geometries to '{trj_fn}'.")

        group_name = f"scan_{title}"
        # with h5py.File("scan.h5", "a") as handle:
        #     try:
        #         del handle[group_name]
        #     except KeyError:
        #         pass
        #     group = handle.create_group(group_name)
        #     group.create_dataset("energies", data=scan_energies)
        #     group.create_dataset("cart_coords", data=scan_cart_coords)
        #     group.create_dataset("target_values", data=target_values)
        #     group.create_dataset("actual_values", data=actual_values)
        #     dt = h5py.vlen_dtype(np.dtype("int32"))
        #     cp = np.array([(pt.value, *ind) for pt, *ind in constrain_prims])
        #     cp_dataset = group.create_dataset(
        #         "constrain_prims", (len(target_values),), dtype=dt
        #     )
        #     cp_dataset[:] = cp
    return scan_geom, scan_cart_coords, scan_energies


def relaxed_1d_scan(
    geom,
    calc_getter,
    constrain_prims,
    start,
    step_size,
    steps,
    opt_key,
    opt_kwargs,
    pref=None,
    callback=None,
):
    assert geom.coord_type == "redund", "'coord_type: redund' is required!"
    if pref is None:
        pref = ""
    else:
        pref = f"{pref}_"

    constr_prim = constrain_prims[0]
    copy_kwargs = {
        "coord_type": "redund",
        "coord_kwargs": {
            "constrain_prims": constrain_prims,
            "recalc_B": True,
        },
    }
    constr_geom = geom.copy(**copy_kwargs)
    constr_ind = constr_geom.internal.get_index_of_typed_prim(constrain_prims[0])
    calc = calc_getter()
    # Always return same calculator, so orbitals can be reused etc.
    def _calc_getter():
        return calc

    # Drop PrimType at index 0
    unit = "au" if len(constr_prim[1:]) == 2 else "rad"

    org_val = constr_geom.coords[constr_ind]
    cur_val = start
    end_val = start + step_size * steps
    target_scan_vals = np.linspace(cur_val, end_val, steps + 1)
    print(
        f"    Coordinate: {constr_prim}\n"
        f"Original value: {org_val:.4f} {unit}\n"
        f" Initial value: {cur_val:.4f} {unit}\n"
        f"   Final value: {end_val:.4f} {unit}\n"
        f"         Steps: {steps}+1\n"
        f"     Step size: {step_size:.4f} {unit}\n"
    )
    scan_vals = list()
    scan_geoms = list()
    scan_energies = list()
    scan_xyzs = list()

    for cycle, cur_val in enumerate(target_scan_vals):
        opt_kwargs_ = opt_kwargs.copy()
        name = f"{pref}relaxed_scan_{cycle:04d}"
        opt_kwargs_["prefix"] = name
        opt_kwargs_["h5_group_name"] = name
        # Keep a copy
        scan_geoms.append(constr_geom.copy())

        # Update constrained coordinate
        new_coords = constr_geom.coords
        new_coords[constr_ind] = cur_val
        constr_geom.set_coords(new_coords, update_constraints=True)

        title = f"{pref}Step {cycle:02d}, coord={cur_val:.4f} {unit}"
        opt_result = run_opt(
            constr_geom, _calc_getter, opt_key, opt_kwargs_, title=title, level=1
        )
        if callback is not None:
            callback(opt_result)
        scan_xyzs.append(constr_geom.as_xyz())
        scan_vals.append(constr_geom.coords[constr_ind])
        scan_energies.append(constr_geom.energy)

        if not opt_result.opt.is_converged:
            print(f"Step {cycle} did not converge. Breaking!")
            break

    with open(f"{pref}relaxed_scan_trj.xyz", "w") as handle:
        handle.write("\n".join(scan_xyzs))

    scan_data = np.stack((scan_vals, scan_energies), axis=1)
    np.savetxt(f"{pref}relaxed_scan.dat", scan_data)
    scan_vals, scan_energies = scan_data.T

    print(highlight_text(f"Scan summary", level=1))
    col_fmts = "int float float float float".split()
    header = ("Step", f"Target / {unit}", f"Actual / {unit}", f"Δ / {unit}", "E / au")
    table = TablePrinter(header, col_fmts, width=14)
    table.print_header()
    for step, (target, act, en) in enumerate(
        zip(target_scan_vals, scan_vals, scan_energies)
    ):
        delta = target - act
        table.print_row((step, target, act, delta, en))
    return scan_geoms, scan_vals, scan_energies



try:
    import psi4
except ImportError:
    pass
import numpy as np

from pysisyphus.calculators.Calculator import Calculator
from pysisyphus.constants import BOHR2ANG
from pysisyphus.xyzloader import make_xyz_str


class PyPsi4(Calculator):
    def __init__(self, method, basis, **kwargs):
        super().__init__(**kwargs)

        self.method = method
        self.basis = basis

        self.meth_bas = f"{self.method}/{self.basis}"

        psi4.core.be_quiet()
        psi4.set_memory(f"{self.pal*self.mem} MB")

    def get_forces(self, atoms, coords):
        xyz = make_xyz_str(atoms, coords.reshape(-1, 3) * BOHR2ANG)
        mol = psi4.geometry(xyz)
        energy, wfn = psi4.energy(self.meth_bas, return_wfn=True)
        gradient = psi4.gradient(self.meth_bas, molecule=mol, ref_wfn=wfn)

        results = {
            "energy": energy,
            "forces": -np.asarray(gradient).flatten(),
        }
        return results

    def __str__(self):
        return f"PyPsi4({self.name})"



import numpy as np

from pysisyphus.intcoords.Primitive import Primitive


class Cartesian(Primitive):

    def __init__(self, *args, **kwargs):
        kwargs["calc_kwargs"] = ("cart_axis",)
        super().__init__(*args, **kwargs)

    @staticmethod
    def _weight(atoms, coords3d, indices, f_damping):
        return 1

    @staticmethod
    def _calculate(coords3d, indices, gradient=False, cart_axis=0):
        m,  = indices
        value = coords3d[m, cart_axis]
        if gradient:
            row = np.zeros_like(coords3d)
            row[m, cart_axis] = 1
            row = row.flatten()
            return value, row
        return value

    @staticmethod
    def _jacobian(coords3d, indices, cart_axis):
        return np.zeros(1)


class CartesianX(Cartesian):
    cart_axis = 0


class CartesianY(Cartesian):
    cart_axis = 1


class CartesianZ(Cartesian):
    cart_axis = 2



# [1] https://doi.org/10.1007/s002140050387
#     Bofill, 1998


import numpy as np

from pysisyphus.tsoptimizers.TSHessianOptimizer import TSHessianOptimizer

import torch

class RSIRFOptimizer(TSHessianOptimizer):
    def optimize(self):
        energy, gradient, H, eigvals, eigvecs, resetted = self.housekeeping()
        self.update_ts_mode(eigvals, eigvecs)

        self.log(
            "Using projection to construct image potential gradient "
            f"and hessian for root(s) {self.roots}."
        )
        # Ensure gradient is in the same subspace as eigvecs (active DOFs)
        if isinstance(H, torch.Tensor):
            if gradient.size(0) != eigvecs.size(0):
                gradient = self.active_from_full(gradient)
        else:
            if gradient.size != eigvecs.shape[0]:
                gradient = self.active_from_full(gradient)
        # Projection matrix to construct g* and H*
        if isinstance(H, torch.Tensor):
            P = torch.eye(gradient.size(0), device=H.device, dtype=H.dtype)
            for root in self.roots:
                trans_vec = eigvecs[:, root]
                P -= 2 * torch.outer(trans_vec, trans_vec)
            H_star = P @ H
            eigvals_, eigvecs_ = torch.linalg.eigh(H_star)
        else:
            P = np.eye(gradient.size)
            for root in self.roots:
                trans_vec = eigvecs[:, root]
                P -= 2 * np.outer(trans_vec, trans_vec)
            H_star = P.dot(H)
            eigvals_, eigvecs_ = np.linalg.eigh(H_star)
        # Neglect small eigenvalues
        eigvals_, eigvecs_ = self.filter_small_eigvals(eigvals_, eigvecs_)

        if isinstance(H, torch.Tensor):
            grad_star = P @ gradient
        else:
            grad_star = P.dot(gradient)
        step = self.get_rs_step(eigvals_, eigvecs_, grad_star, name="RS-I-RFO")

        self.predicted_energy_changes.append(self.rfo_model(gradient, self.cur_H, step))

        step = self.full_from_active(step)
        if isinstance(step, torch.Tensor):
            step = step.cpu().numpy()
        return step



import logging
import sys

from pysisyphus.stocastic.Kick import Kick
from pysisyphus.stocastic.FragmentKick import FragmentKick


__all__ = [
    "FragmentKick",
    "Kick",
]

logger = logging.getLogger("stocastic")
logger.setLevel(logging.DEBUG)
fh = logging.FileHandler("stocastic.log", mode="w", delay=True)
fh.setLevel(logging.DEBUG)
logger.addHandler(logging.StreamHandler(sys.stdout))
# fmt_str = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
# formatter = logging.Formatter(fmt_str)
# fh.setFormatter(formatter)
logger.addHandler(fh)



import numpy as np

def interpol_alpha_quad(f_0, df_0, f_alpha_0, alpha_0):
    return -df_0*alpha_0**2 / 2 / (f_alpha_0 - f_0 - df_0*alpha_0)


def interpol_alpha_cubic(f_0, df_0, f_alpha_0, f_alpha_1, alpha_0, alpha_1):
    quot = 1 / (alpha_0**2 * alpha_1**2 * (alpha_1 - alpha_0))
    A = np.array(((alpha_0**2, -alpha_1**2),
                  (-alpha_0**3, alpha_1**3)))
    B = np.array(((f_alpha_1 - f_0 - df_0*alpha_1),
                  (f_alpha_0 - f_0 - df_0*alpha_0)))
    a, b = quot * A @ B
    alpha_cubic = (-b + (b**2 - 3*a*df_0)**(0.5)) / (3*a)
    return alpha_cubic



"""Unit tests for pdb2reaction.extract pure helper functions."""

from __future__ import annotations

import pytest


class TestFormatEchoMessage:
    def test_no_args(self):
        from pdb2reaction.extract import _format_echo_message

        assert _format_echo_message("hello") == "hello"

    def test_with_args(self):
        from pdb2reaction.extract import _format_echo_message

        assert _format_echo_message("value=%d", 42) == "value=42"

    def test_format_error_fallback(self):
        from pdb2reaction.extract import _format_echo_message

        # Wrong format specifier should fall back gracefully
        result = _format_echo_message("value=%d", "not_a_number")
        assert isinstance(result, str)


class TestParseResTokens:
    def test_single_resseq(self):
        from pdb2reaction.extract import _parse_res_tokens

        result = _parse_res_tokens("123")
        assert len(result) == 1
        chain, resseq, icode = result[0]
        assert resseq == 123

    def test_chain_resseq(self):
        from pdb2reaction.extract import _parse_res_tokens

        result = _parse_res_tokens("A:123")
        assert len(result) == 1
        chain, resseq, icode = result[0]
        assert chain == "A"
        assert resseq == 123

    def test_chain_resseq_icode(self):
        from pdb2reaction.extract import _parse_res_tokens

        result = _parse_res_tokens("A:123A")
        assert len(result) == 1
        chain, resseq, icode = result[0]
        assert chain == "A"
        assert resseq == 123
        assert icode == "A"

    def test_multiple(self):
        from pdb2reaction.extract import _parse_res_tokens

        result = _parse_res_tokens("A:10,B:20")
        assert len(result) == 2


class TestMaxSerialFromPdbText:
    def test_basic(self):
        from pdb2reaction.extract import _max_serial_from_pdb_text

        pdb_text = (
            "ATOM      1  N   ALA A   1       1.000   2.000   3.000  1.00  0.00           N\n"
            "ATOM     42  CA  ALA A   1       4.000   5.000   6.000  1.00  0.00           C\n"
        )
        assert _max_serial_from_pdb_text(pdb_text) == 42

    def test_empty(self):
        from pdb2reaction.extract import _max_serial_from_pdb_text

        assert _max_serial_from_pdb_text("") == 0


class TestStripTrailingEND:
    def test_strips_end(self):
        from pdb2reaction.extract import _strip_trailing_END

        text = "ATOM      1  N   ALA A   1\nEND\n"
        result = _strip_trailing_END(text)
        assert "END" not in result.split("\n")[-1] or result.strip().endswith("ALA A   1")

    def test_no_end(self):
        from pdb2reaction.extract import _strip_trailing_END

        text = "ATOM      1  N   ALA A   1\n"
        result = _strip_trailing_END(text)
        assert result == text


class TestParseLigandChargeOption:
    def test_none(self):
        from pdb2reaction.extract import _parse_ligand_charge_option

        total, mapping = _parse_ligand_charge_option(None)
        assert total is None
        assert mapping is None

    def test_numeric_int(self):
        from pdb2reaction.extract import _parse_ligand_charge_option

        total, mapping = _parse_ligand_charge_option(-3)
        assert total == -3.0
        assert mapping is None

    def test_numeric_string(self):
        from pdb2reaction.extract import _parse_ligand_charge_option

        total, mapping = _parse_ligand_charge_option("-3")
        assert total == -3.0
        assert mapping is None

    def test_mapping_string(self):
        from pdb2reaction.extract import _parse_ligand_charge_option

        total, mapping = _parse_ligand_charge_option("GPP:-3,SAM:1")
        assert total is None
        assert mapping == {"GPP": -3.0, "SAM": 1.0}

    def test_dict_input(self):
        from pdb2reaction.extract import _parse_ligand_charge_option

        total, mapping = _parse_ligand_charge_option({"GPP": -3, "SAM": 1})
        assert total is None
        assert mapping == {"GPP": -3.0, "SAM": 1.0}


class TestFormatLinkHBlock:
    def test_basic(self):
        from pdb2reaction.extract import _format_linkH_block

        coords = [(1.0, 2.0, 3.0), (4.0, 5.0, 6.0)]
        result = _format_linkH_block(coords, start_serial=100)
        lines = [l for l in result.strip().split("\n") if l.strip()]
        assert len(lines) == 2
        assert "HETATM" in lines[0]
        # start_serial=100 means first atom gets serial 101
        assert "101" in lines[0] or " 101" in lines[0]



#!/usr/bin/env python3
"""
fix_altloc.py - Drop alternate locations from PDB files

What it does
------------
1) Blank the PDB altLoc column (column 17, 1-based) with a single space.
   - This is a 1-character replacement (no shifting / no reformatting).
2) If the same atom appears multiple times due to alternate locations
   (altLoc like A/B/... or custom labels like H/L),
   keep the "best" one by the default rule:
      - Highest occupancy first
      - If tied (or occupancy missing), keep the earliest one in the file

Handled records
---------------
- ATOM / HETATM
- ANISOU is also handled: ANISOU lines are kept only if the corresponding
  ATOM/HETATM line (same serial) is kept.

Notes
-----
- Atom serial numbers are NOT renumbered (gaps may remain).
- CONECT and other connectivity/annotation records are NOT updated.

Usage
-----
  pdb2reaction fix-altloc -i input.pdb -o output.pdb
  pdb2reaction fix-altloc -i ./dir -o ./dir_clean --recursive
  pdb2reaction fix-altloc -i ./dir --inplace --recursive
"""

import shutil
from pathlib import Path
from typing import Dict, Iterable, Iterator, List, Optional, Set, Tuple

import click

COORD_RECORDS = ("ATOM  ", "HETATM")
ANISOU_RECORD = "ANISOU"

# PDB fixed columns (0-based Python indices)
ALTLOC_IDX = 16                 # column 17 (1-based)
SERIAL_SLICE = slice(6, 11)     # columns 7-11 (1-based), width 5
OCC_SLICE = slice(54, 60)       # columns 55-60 (1-based), width 6


def split_newline(line: str) -> Tuple[str, str]:
    """Split a line into (core, newline) while preserving the newline exactly."""
    if line.endswith("\r\n"):
        return line[:-2], "\r\n"
    if line.endswith("\n"):
        return line[:-1], "\n"
    if line.endswith("\r"):
        return line[:-1], "\r"
    return line, ""


def ensure_len(core: str, n: int) -> str:
    """Right-pad with spaces to guarantee at least n characters (no shifting)."""
    return core if len(core) >= n else core.ljust(n)


def blank_altloc(line: str) -> str:
    """
    Blank the altLoc field (column 17, 1-based) with a single space.

    IMPORTANT: This does NOT remove characters; it replaces exactly one character,
    so the fixed-width PDB formatting is preserved.
    """
    core, nl = split_newline(line)
    core = ensure_len(core, ALTLOC_IDX + 1)  # make sure core[ALTLOC_IDX] exists
    core = core[:ALTLOC_IDX] + " " + core[ALTLOC_IDX + 1:]
    return core + nl


def atom_serial_5(line: str) -> str:
    """Return the 5-character atom serial field exactly as it appears (cols 7-11)."""
    core, _ = split_newline(line)
    core = ensure_len(core, SERIAL_SLICE.stop)
    return core[SERIAL_SLICE]


def parse_occupancy(line: str) -> Optional[float]:
    """
    Parse occupancy from columns 55-60 (1-based).
    Returns None if missing/unparseable.
    """
    core, _ = split_newline(line)
    core = ensure_len(core, OCC_SLICE.stop)
    s = core[OCC_SLICE].strip()
    if not s:
        return None
    try:
        return float(s)
    except ValueError:
        return None


def atom_identity_key(line: str) -> Tuple[str, str, str, str, str, str, str]:
    """
    Build a key to identify the "same atom" while IGNORING altLoc.

    Fields used (fixed columns, classic PDB):
      - record name (ATOM/HETATM)     cols 1-6
      - atom name                      cols 13-16
      - residue name                   cols 18-20
      - chain ID                       col 22
      - residue sequence number        cols 23-26
      - insertion code                 col 27
      - segID (non-standard, common)   cols 73-76

    segID is included to reduce accidental merging in MD-style PDBs where chain ID may be blank.
    """
    core, _ = split_newline(line)
    core = ensure_len(core, 76)

    record = core[0:6]
    atom_name = core[12:16]
    res_name = core[17:20]
    chain_id = core[21:22]
    res_seq = core[22:26]
    i_code = core[26:27]
    seg_id = core[72:76]  # 73-76 (1-based), optional/non-standard

    return (record, atom_name, res_name, chain_id, res_seq, i_code, seg_id)


def process_block(lines: List[str]) -> List[str]:
    """
    Two-pass processing for a block (either the whole file if no MODEL,
    or the content between MODEL and ENDMDL):

    Pass 1: determine the best coordinate line per atom key
            by (occupancy desc, first-appearance asc).
    Pass 2: output only the chosen coordinate lines (altLoc blanked),
            and keep only ANISOU lines whose serial is chosen (altLoc blanked).
            All other records are passed through unchanged.

    Handling of different atom counts between altLoc states:
    --------------------------------------------------------
    When different altLoc states have different atoms (e.g., altLoc A has
    atoms N,CA,CB,CG while altLoc B has N,CA,CB,CD), this function:
    - For DUPLICATE atoms (same identity key, e.g., N,CA,CB): selects the best
      one based on occupancy
    - For UNIQUE atoms (only in one altLoc, e.g., CG in A, CD in B): keeps ALL
      of them in the output

    This ensures the output structure contains all unique atoms from all altLoc
    states, with duplicates resolved to the best conformer.
    """
    # key -> (occ_val_for_compare, line_index, serial5)
    best: Dict[Tuple[str, str, str, str, str, str, str], Tuple[float, int, str]] = {}

    for idx, line in enumerate(lines):
        if line.startswith(COORD_RECORDS):
            key = atom_identity_key(line)
            occ = parse_occupancy(line)
            occ_val = occ if occ is not None else float("-inf")
            serial = atom_serial_5(line)

            if key not in best:
                best[key] = (occ_val, idx, serial)
            else:
                best_occ, best_idx, _best_serial = best[key]
                # Prefer higher occupancy; if tied, prefer earlier line (smaller idx)
                if (occ_val > best_occ) or (occ_val == best_occ and idx < best_idx):
                    best[key] = (occ_val, idx, serial)

    chosen_serials: Set[str] = set(v[2] for v in best.values())

    out: List[str] = []
    for idx, line in enumerate(lines):
        if line.startswith(COORD_RECORDS):
            key = atom_identity_key(line)
            # Keep only the selected "best" line for this key
            if key in best and best[key][1] == idx:
                out.append(blank_altloc(line))
            continue

        if line.startswith(ANISOU_RECORD):
            serial = atom_serial_5(line)
            if serial in chosen_serials:
                out.append(blank_altloc(line))
            continue

        out.append(line)

    return out


def process_stream(lines: Iterable[str]) -> Iterator[str]:
    """
    Handle MODEL/ENDMDL blocks:
      - If MODEL records exist, apply the selection independently within each MODEL block.
      - Text outside MODEL blocks is processed as a single block.
    """
    buffer: List[str] = []
    in_model = False

    for line in lines:
        if line.startswith("MODEL "):
            # Flush anything accumulated before this MODEL
            if buffer:
                for x in process_block(buffer):
                    yield x
                buffer = []
            in_model = True
            yield line
            continue

        if in_model and line.startswith("ENDMDL"):
            # Process the model contents, then emit ENDMDL
            for x in process_block(buffer):
                yield x
            buffer = []
            in_model = False
            yield line
            continue

        buffer.append(line)

    # Flush remaining lines at EOF
    if buffer:
        for x in process_block(buffer):
            yield x


def clean_pdb_file(in_path: Path, out_path: Path) -> None:
    """Process a PDB file and write the cleaned output."""
    out_path.parent.mkdir(parents=True, exist_ok=True)
    with in_path.open("r", newline="") as fin, out_path.open("w", newline="") as fout:
        for out_line in process_stream(fin):
            fout.write(out_line)


def collect_pdb_files(input_path: Path, recursive: bool) -> List[Path]:
    """Collect *.pdb files from a file or directory (optionally recursive)."""
    if input_path.is_file():
        return [input_path]
    pattern = "**/*.pdb" if recursive else "*.pdb"
    return sorted([p for p in input_path.glob(pattern) if p.is_file()])


# =============================================================================
# Public API for programmatic use
# =============================================================================

def has_altloc(pdb_path: Path) -> bool:
    """
    Check if a PDB file contains any non-blank altLoc characters (column 17, 1-based).

    Returns True if at least one ATOM/HETATM record has a non-space character
    in the altLoc column. Returns False if no altLoc is found.
    """
    try:
        with open(pdb_path, "r", encoding="utf-8", errors="ignore") as fh:
            for line in fh:
                if line.startswith(COORD_RECORDS):
                    # altLoc is at column 17 (1-based), which is index 16 (0-based)
                    if len(line) > ALTLOC_IDX:
                        altloc_char = line[ALTLOC_IDX]
                        if altloc_char != " " and altloc_char != "":
                            return True
        return False
    except Exception:
        return False


def fix_altloc_file(
    in_path: str | Path,
    out_path: str | Path,
    *,
    overwrite: bool = False,
    skip_if_no_altloc: bool = True,
) -> bool:
    """
    Fix alternate locations in a PDB file.

    Parameters
    ----------
    in_path : str | Path
        Input PDB file path.
    out_path : str | Path
        Output PDB file path.
    overwrite : bool
        If True, overwrite existing output file. Default False.
    skip_if_no_altloc : bool
        If True, skip processing if no altLoc is detected. Default True.

    Returns
    -------
    bool
        True if the file was processed (altloc found and fixed),
        False if skipped (no altloc detected).

    Raises
    ------
    FileExistsError
        If output file exists and overwrite=False.
    FileNotFoundError
        If input file does not exist.
    """
    in_path = Path(in_path)
    out_path = Path(out_path)

    if not in_path.exists():
        raise FileNotFoundError(f"Input file not found: {in_path}")

    if out_path.exists() and not overwrite:
        raise FileExistsError(f"Output file exists: {out_path}")

    if skip_if_no_altloc and not has_altloc(in_path):
        return False

    clean_pdb_file(in_path, out_path)
    return True


# =============================================================================
# CLI
# =============================================================================

def _run_fix_altloc(
    input_path: Path,
    out: Optional[Path],
    recursive: bool,
    inplace: bool,
    overwrite: bool,
    force: bool,
) -> None:
    """Business logic for fix-altloc (separated from CLI layer)."""
    pdb_files = collect_pdb_files(input_path, recursive)
    if not pdb_files:
        raise click.ClickException(f"No .pdb files found in: {input_path}")

    skip_if_no_altloc = not force
    processed_count = 0
    skipped_count = 0

    # In-place mode
    if inplace:
        for in_path in pdb_files:
            if skip_if_no_altloc and not has_altloc(in_path):
                skipped_count += 1
                continue

            bak_path = in_path.with_suffix(in_path.suffix + ".bak")
            if not bak_path.exists():
                shutil.copy2(in_path, bak_path)

            tmp_path = in_path.with_suffix(in_path.suffix + ".tmp")
            clean_pdb_file(in_path, tmp_path)
            tmp_path.replace(in_path)
            processed_count += 1

        if processed_count > 0:
            click.echo(f"[fix-altloc] Processed {processed_count} file(s) in-place.")
        if skipped_count > 0:
            click.echo(f"[fix-altloc] Skipped {skipped_count} file(s) (no altLoc detected).")
        return

    # File input
    if input_path.is_file():
        in_path = input_path

        if skip_if_no_altloc and not has_altloc(in_path):
            click.echo(f"[fix-altloc] Skipped {in_path} (no altLoc detected).")
            return

        if out is None:
            out_path = in_path.with_name(in_path.stem + "_clean.pdb")
        else:
            if out.suffix.lower() == ".pdb":
                out_path = out
            else:
                out.mkdir(parents=True, exist_ok=True)
                out_path = out / in_path.name

        if out_path.exists() and not overwrite:
            raise click.ClickException(
                f"Output exists: {out_path} (use --overwrite to overwrite)"
            )

        clean_pdb_file(in_path, out_path)
        click.echo(f"[fix-altloc] Fixed altLoc → {out_path}")
        return

    # Directory input
    in_dir = input_path
    out_dir = out if out is not None else in_dir.with_name(in_dir.name + "_clean")
    out_dir.mkdir(parents=True, exist_ok=True)

    for in_path in pdb_files:
        if skip_if_no_altloc and not has_altloc(in_path):
            skipped_count += 1
            continue

        rel = in_path.relative_to(in_dir)
        out_path = out_dir / rel

        if out_path.exists() and not overwrite:
            raise click.ClickException(
                f"Output exists: {out_path} (use --overwrite to overwrite)"
            )

        clean_pdb_file(in_path, out_path)
        processed_count += 1

    if processed_count > 0:
        click.echo(f"[fix-altloc] Processed {processed_count} file(s) → {out_dir}")
    if skipped_count > 0:
        click.echo(f"[fix-altloc] Skipped {skipped_count} file(s) (no altLoc detected).")


@click.command(
    name="fix-altloc",
    help=(
        "Blank PDB altLoc column (col 17) without shifting, and keep one altLoc "
        "per atom by default rule: highest occupancy, then earliest appearance."
    ),
    context_settings={"help_option_names": ["-h", "--help"]},
)
@click.option(
    "-i", "--input", "input_path",
    type=click.Path(exists=True, path_type=Path),
    required=True,
    help="Input PDB file or directory.",
)
@click.option(
    "-o", "--out",
    type=click.Path(path_type=Path),
    default=None,
    help="Output file (if input is a file) or output directory (if input is a directory).",
)
@click.option(
    "--recursive/--no-recursive",
    default=False,
    show_default=True,
    help="When input is a directory, process *.pdb recursively (including subdirectories).",
)
@click.option(
    "--inplace/--no-inplace",
    default=False,
    show_default=True,
    help="Overwrite input file(s) in place (creates .bak next to each file).",
)
@click.option(
    "--overwrite/--no-overwrite",
    default=False,
    show_default=True,
    help="Allow overwriting existing output files.",
)
@click.option(
    "--force/--no-force",
    default=False,
    show_default=True,
    help="Process files even if no altLoc is detected (default: skip files without altLoc).",
)
def cli(
    input_path: Path,
    out: Optional[Path],
    recursive: bool,
    inplace: bool,
    overwrite: bool,
    force: bool,
) -> None:
    _run_fix_altloc(
        input_path=input_path,
        out=out,
        recursive=recursive,
        inplace=inplace,
        overwrite=overwrite,
        force=force,
    )


if __name__ == "__main__":
    cli()



# [1] https://aip.scitation.org/doi/pdf/10.1063/5.0004046
#     Efficient implementation of the superposition of atomic potentials initial
#     guess for electronic structure calculations in Gaussian basis sets
#     Lehtola, Visscher, Engel,  2020

import itertools as it
from math import sqrt, log, pi
from pathlib import Path
import textwrap
from typing import List, Literal


import h5py
from jinja2 import Template
from joblib import Memory
import numpy as np
from numpy.typing import NDArray
import scipy as sp


from pysisyphus.config import L_MAX, L_AUX_MAX
from pysisyphus.elem_data import (
    ATOMIC_NUMBERS,
    INV_ATOMIC_NUMBERS,
    nuc_charges_for_atoms,
)
from pysisyphus.helpers_pure import file_or_str
from pysisyphus.linalg import multi_component_sym_mat
from pysisyphus.wavefunction.helpers import (
    canonical_order,
    get_l,
    get_shell_shape,
    L_MAP_INV,
    permut_for_order,
)

from pysisyphus.wavefunction.ints import (
    cart_gto3d,
    coulomb3d,
    diag_quadrupole3d,
    dipole3d,
    kinetic3d,
    ovlp3d,
    quadrupole3d,
    int2c2e3d,
    int3c2e3d_sph,
)

from pysisyphus.wavefunction.cart2sph import cart2sph_coeffs
from pysisyphus.wavefunction.normalization import norm_cgto_lmn


class Shell:
    def __init__(
        self,
        L: int,
        center: NDArray[float],
        coeffs: NDArray[float],
        exps: NDArray[float],
        center_ind: int,
        atomic_num=None,
        shell_index=None,
        index=None,
        # min_coeff: float = 1e-8,
    ):
        self.L = get_l(L)
        self.center = np.array(center, dtype=float)  # (x, y, z), 1d array
        coeffs = np.array(coeffs, dtype=float)
        # cur_min_coeff = np.abs(coeffs).min()
        # assert (
        # cur_min_coeff >= min_coeff
        # ), f"Got invalid contraction coeff of '{cur_min_coeff:.6f}'"
        # Store original contraction coefficients
        self.coeffs_org = coeffs.copy()
        # Orbital exponents, 1d array
        self.exps = np.array(exps, dtype=float)
        assert self.coeffs_org.size == self.exps.size
        coeffs = np.array(coeffs)
        assert coeffs.size == self.exps.size
        self.center_ind = int(center_ind)
        if atomic_num is None:
            atomic_num = -1
        self.atomic_num = int(atomic_num)
        if shell_index is not None:
            self.shell_index = shell_index
        if index is not None:
            self.index = index

        self.coeffs, _ = norm_cgto_lmn(coeffs, self.exps, self.L)

        try:
            self.atom = INV_ATOMIC_NUMBERS[self.atomic_num]
        except KeyError:
            self.atom = "X"
        self.atom = self.atom.lower()

    def as_tuple(self):
        return self.L, self.center, self.coeffs, self.exps, self.index, self.size

    def exps_coeffs_iter(self):
        return zip(self.exps, self.coeffs)

    @property
    def contr_depth(self):
        return self.coeffs.size

    @property
    def cart_powers(self):
        return np.array(canonical_order(self.L), dtype=int)

    @property
    def size(self):
        """Number of cartesian basis functions in the shell."""
        return self.cart_size

    @property
    def cart_size(self):
        """Number of cartesian basis functions in the shell."""
        L = self.L
        return (L + 1) * (L + 2) // 2

    @property
    def sph_size(self):
        """Number of cartesian basis functions in the shell."""
        return 2 * self.L + 1

    @property
    def index(self) -> int:
        return self._index

    @index.setter
    def index(self, index: int):
        self._index = int(index)

    @property
    def shell_index(self) -> int:
        return self._shell_index

    @shell_index.setter
    def shell_index(self, shell_index: int):
        self._shell_index = shell_index

    def to_pyscf_shell(self):
        key = L_MAP_INV[self.L]
        lines = [
            f"{self.atom.title()}    {key.upper()}",
        ]
        lines += [
            f"{exp_:> 18.10f}    {coeff:>18.10f}"
            for exp_, coeff in zip(self.exps, self.coeffs)
        ]
        return "\n".join(lines)

    def to_sap_shell(self):
        """Fix contraction coefficients, for use in SAP-initial guess calculation.

        See [1]."""
        # SAP shells must always be s-shells
        assert self.L == 0
        self.coeffs = self.coeffs_org
        coeffs_sum = self.coeffs.sum()
        # Check sum rule; sum should equal -Z.
        assert (coeffs_sum + ATOMIC_NUMBERS[self.atom.lower()]) <= 1e-5
        # The potential fits were carried out for functions of the form:
        #   g_p(r) = (α_p / π)**1.5 * exp(-α_p / r) .
        # So we multiply the prefactor onto the contraction coefficients.
        # See [1], p. 152, just above Eq. (2).
        self.coeffs *= (self.exps / np.pi) ** 1.5

    def __str__(self):
        try:
            center_str = f", at atom {self.center_ind}"
        except AttributeError:
            center_str = ""
        return f"Shell(L={self.L}, {self.contr_depth} pGTO{center_str})"

    def __repr__(self):
        return self.__str__()


Ordering = Literal["native", "pysis"]


class Shells:
    sph_Ps = {l: np.eye(2 * l + 1) for l in range(max(L_MAX, L_AUX_MAX) + 1)}

    def __init__(
        self,
        shells: List[Shell],
        screen: bool = False,
        cache: bool = False,
        cache_path: str = "./cache",
        ordering: Ordering = "native",
    ):
        self.shells = shells
        self.ordering = ordering
        self.screen = screen
        self.cache = cache
        self.cache_path = Path(cache_path)
        """
        'native' refers to the original ordering, as used in the QC program. The
        ordering will be reflected in the MO coefficient ordering. With 'native'
        the integrals calculated by pysisyphus must be reorderd, to match the native
        ordering of the MO coefficients.
        """
        assert ordering in ("pysis", "native")

        # Now that we have all Shell objects, we can set their starting indices
        index = 0
        shell_index = 0
        for shell in self.shells:
            shell.shell_index = shell_index
            shell.index = index
            index += shell.size
            shell_index += 1

        # Try to construct Cartesian permutation matrix from cart_order, if defnied.
        try:
            self.cart_Ps = permut_for_order(self.cart_order)
        except AttributeError:
            pass

        self.atoms, self.coords3d = self.atoms_coords3d
        # Precontract & store coefficients for reordering spherical basis functions
        # and converting them from Cartesian basis functions.
        self.reorder_c2s_coeffs = self.P_sph @ self.cart2sph_coeffs

        # Enable disk cache for 1el-integrals
        if self.cache:
            self.memory = Memory(self.cache_path, verbose=0)
            self.get_1el_ints_cart = self.memory.cache(self.get_1el_ints_cart)
            self.get_1el_ints_sph = self.memory.cache(self.get_1el_ints_sph)

    def __len__(self):
        return len(self.shells)

    def __getitem__(self, key):
        return self.shells[key]

    def print_shells(self):
        for shell in self.shells:
            print(shell)

    def as_tuple(self):
        # Ls, centers, contr_coeffs, exponents, indices, sizes
        return zip(*[shell.as_tuple() for shell in self.shells])

    @property
    def cart_size(self):
        """Number of cartesian basis functions."""
        return sum([shell.cart_size for shell in self.shells])

    @property
    def sph_size(self):
        """Number of spherical basis."""
        return sum([shell.sph_size for shell in self.shells])

    def as_arrays(self, fortran=False):
        """Similar to the approach in libcint."""
        # bas_centers
        # One entry per shell, integer array.
        #   center_ind, atomic number, pointer to center coordinates in bas_data (3 integers)
        bas_centers = list()
        # bas_spec
        # One entry per shell, integer array.
        #   shell_ind, total angmom, N_pgto, N_cgto, \
        #   pointer to contraction coefficients and exponents in bas_data \
        #   (2*N_pgto floats)
        bas_spec = list()
        # bas_data
        # Float array. Starts with 3 * N_centers floats, containing the center
        # coordinates. Continues with alternating contraction coefficient and
        # orbital exponent data.
        bas_data = list()
        pointer = 1 if fortran else 0

        shells = self.shells
        # Store center data, i.e., where the basis functions are located.
        for shell in shells:
            center_ind = shell.center_ind
            bas_data.extend(shell.center)  # Store coordinates
            atomic_num = shell.atomic_num
            if atomic_num is None:
                atomic_num = -1
            bas_centers.append((center_ind, atomic_num, pointer))
            pointer += 3

        # Store contraction coefficients and primitive exponents.
        for shell_ind, shell in enumerate(self.shells):
            # L, center, coeffs, exponents, index, size = shell.as_tuple()
            L, _, _, exponents, _, _ = shell.as_tuple()
            contr_coeffs = shell.coeffs_org
            nprim = len(contr_coeffs)
            # ncontr is hardcoded to 1 for now, as there are no special functions
            # to handle general contractions.
            ncontr = 1
            bas_spec.append((shell_ind, L, nprim, ncontr, pointer))
            bas_data.extend(contr_coeffs)
            pointer += nprim
            bas_data.extend(exponents)
            pointer += nprim

        bas_centers = np.array(bas_centers, dtype=int)
        bas_spec = np.array(bas_spec, dtype=int)
        bas_data = np.array(bas_data, dtype=float)
        return bas_centers, bas_spec, bas_data

    def to_sympleints_arrays(self):
        shells = self.shells
        # center_ind, atomic_num, L, nprims
        centers = np.zeros((len(shells), 3))
        shell_data = list()
        coefficients = list()
        exponents = list()
        for i, shell in enumerate(shells):
            coeffs = shell.coeffs_org
            coefficients.extend(coeffs.tolist())
            exponents.extend(shell.exps.tolist())
            nprims = len(coeffs)

            centers[i] = shell.center
            atomic_num = shell.atomic_num
            if atomic_num is None:
                atomic_num = -1
            shell_data.append((shell.center_ind, atomic_num, shell.L, nprims))
        shell_data = np.array(shell_data, dtype=np.int32)
        coefficients = np.array(coefficients)
        exponents = np.array(exponents)
        return shell_data, centers, coefficients, exponents

    def dump_to_h5_group(self, h5_handle, group_name):
        group = h5_handle.create_group(group_name)
        shell_data, centers, coefficients, exponents = self.to_sympleints_arrays()
        group.create_dataset("shell_data", data=shell_data)
        group.create_dataset("centers", data=centers)
        group.create_dataset("coefficients", data=coefficients)
        group.create_dataset("exponents", data=exponents)

    @property
    def l_max(self):
        return max([shell.L for shell in self.shells])

    @property
    def atoms_coords3d(self):
        atoms = list()
        coords3d = list()
        center_inds = list()
        for shell in self.shells:
            center_ind = shell.center_ind
            # Skip cycle if we already registered this center
            if center_ind in center_inds:
                continue
            else:
                center_inds.append(center_ind)
            try:
                atom = INV_ATOMIC_NUMBERS[shell.atomic_num]
            # Use dummy atom when atomic_num is not set / None
            except KeyError:
                atom = "X"
            atoms.append(atom)
            center = shell.center
            coords3d.append(center)
        coords3d = np.array(coords3d)
        return atoms, coords3d

    @property
    def cart_bf_num(self):
        return sum([shell.size for shell in self.shells])

    def from_basis(self, name, shells_cls=None, **kwargs):
        from pysisyphus.wavefunction.Basis import shells_with_basis

        atoms, coords3d = self.atoms_coords3d
        if shells_cls is None:
            shells_cls = type(self)
        return shells_with_basis(
            atoms, coords3d, name=name, shells_cls=shells_cls, **kwargs
        )

    @staticmethod
    @file_or_str(".in")
    def from_aomix(text):
        # import here, to avoid cyclic imports
        from pysisyphus.io.aomix import shells_from_aomix

        shells = shells_from_aomix(text)
        return shells

    @staticmethod
    @file_or_str(".json")
    def from_orca_json(text):
        # import here, to avoid cyclic imports
        from pysisyphus.io.orca import shells_from_json

        shells = shells_from_json(text)
        return shells

    @staticmethod
    @file_or_str(".fchk")
    def from_fchk(text):
        # import here, to avoid cyclic imports
        from pysisyphus.io.fchk import shells_from_fchk

        shells = shells_from_fchk(text)
        return shells

    @staticmethod
    def from_sympleints_arrays(shell_data, centers, coefficients, exponents, **kwargs):
        shells = list()
        pointer = 0
        for i, (center_ind, atomic_num, L, nprims) in enumerate(shell_data):
            coeffs = coefficients[pointer : pointer + nprims]
            exps = exponents[pointer : pointer + nprims]
            shell = Shell(L, centers[i], coeffs, exps, center_ind, atomic_num)
            pointer += nprims
            shells.append(shell)
        return Shells(shells, **kwargs)

    @staticmethod
    def from_sympleints_h5(h5_fn, group_name="shells"):
        with h5py.File(h5_fn, "r") as handle:
            group = handle[group_name]
            shell_data = group["shell_data"][:]
            centers = group["centers"][:]
            coefficients = group["coefficients"][:]
            exponents = group["exponents"][:]
        return Shells.from_sympleints_arrays(
            shell_data, centers, coefficients, exponents
        )

    @staticmethod
    def from_pyscf_mol(mol):
        shells = list()
        for bas_id in range(mol.nbas):
            L = mol.bas_angular(bas_id)
            center = mol.bas_coord(bas_id)
            coeffs = mol.bas_ctr_coeff(bas_id).flatten()
            exps = mol.bas_exp(bas_id)
            assert (
                coeffs.size == exps.size
            ), "General contractions are not yet supported."
            center_ind = mol.bas_atom(bas_id)
            atom_symbol = mol.atom_symbol(center_ind)
            atomic_num = ATOMIC_NUMBERS[atom_symbol.lower()]
            shell = Shell(L, center, coeffs, exps, center_ind, atomic_num)
            shells.append(shell)
        return PySCFShells(shells)

    def to_pyscf_mol(self):
        # TODO: this would be better suited as a method of Wavefunction,
        # as pyscf Moles must have sensible spin & charge etc.
        # TODO: currently, this does not support different basis sets
        # for the same element.
        try:
            from pyscf import gto
        except ModuleNotFoundError:
            return None

        unique_atoms = set()
        basis = dict()
        atoms_coords = list()
        for _, center_shells in self.center_shell_iter():
            center_shells = list(center_shells)
            shell0 = center_shells[0]
            atom = shell0.atom
            x, y, z = shell0.center
            atoms_coords.append(f"{atom} {x} {y} {z}")
            if atom not in unique_atoms:
                unique_atoms.add(atom)
            else:
                continue
            atom_shells = "\n".join([shell.to_pyscf_shell() for shell in center_shells])
            basis[atom] = gto.basis.parse(atom_shells)

        mol = gto.Mole()
        mol.atom = "; ".join(atoms_coords)
        mol.unit = "Bohr"
        mol.basis = basis
        mol.build()
        return mol

    def center_shell_iter(self):
        sorted_shells = sorted(self.shells, key=lambda shell: shell.center_ind)
        return it.groupby(sorted_shells, key=lambda shell: shell.center_ind)

    def fragment_ao_map(self, fragments):
        frag_map = dict()
        for i, frag in enumerate(fragments):
            for center_ind in frag:
                frag_map[center_ind] = i

        frag_ao_map = dict()
        for i, ao_center in enumerate(self.sph_ao_centers):
            frag_ind = frag_map[ao_center]
            frag_ao_map.setdefault(frag_ind, list()).append(i)
        return frag_ao_map

    def as_gau_gbs(self) -> str:
        def dfmt(num):
            return f"{num: 12.10e}".replace("e", "D")

        gbs_tpl = Template(
            """
        {% for center_ind, shells in center_shell_iter %}
            {{ center_ind+1 }} 0
        {%- for shell in shells %}
        {{ ("S", "P", "D", "F", "G")[shell.L] }}   {{ shell.exps.size}}    1.00 0.000000000000
            {%- for exp_, coeff in shell.exps_coeffs_iter() %}
            {{ dfmt(exp_) }} {{ dfmt(coeff) }}
            {%- endfor -%}
        {% endfor %}
         ****
        {%- endfor %}
        """
        )
        rendered = gbs_tpl.render(center_shell_iter=self.center_shell_iter(), dfmt=dfmt)
        rendered = textwrap.dedent(rendered)
        return rendered

    def _ao_center_iter(self, func):
        i = 0
        for shell in self.shells:
            for _ in func(shell.L):
                yield shell.center_ind
                i += 0

    @property
    def cart_ao_centers(self):
        return self._ao_center_iter(canonical_order)

    @property
    def sph_ao_centers(self):
        return self._ao_center_iter(lambda l: range(2 * l + 1))

    @property
    def cart2sph_coeffs(self):
        cart2sph = cart2sph_coeffs(self.l_max)
        C = sp.linalg.block_diag(*[cart2sph[shell.L] for shell in self.shells])
        return C

    @property
    def P_sph(self):
        """Permutation matrix for spherical basis functions."""
        return sp.linalg.block_diag(*[self.sph_Ps[shell.L] for shell in self.shells])

    @property
    def P_cart(self):
        """Permutation matrix for Cartesian basis functions."""
        try:
            P_cart = sp.linalg.block_diag(
                *[self.cart_Ps[shell.L] for shell in self.shells]
            )
        except AttributeError:
            P_cart = np.eye(self.cart_size)
        return P_cart

    def eval(self, xyz, spherical=True):
        """Evaluate all basis functions at points xyz using generated code.

        A possibly more efficient approach is discussed in III C of
        https://doi.org/10.1063/1.469408.
        """
        if spherical:
            precontr = self.cart2sph_coeffs.T @ self.P_sph.T
        else:
            precontr = self.P_cart.T
        ncbfs = precontr.shape[0]
        npoints = len(xyz)

        coords3d = self.coords3d
        grid_vals = np.zeros((npoints, ncbfs))
        for center_ind, shells in self.center_shell_iter():
            # Create list from the shells iterator, otherwise it will be empty
            # after the first pass over all points.
            shells = list(shells)
            for i, R in enumerate(xyz):
                for shell in shells:
                    La, A, da, ax, a_ind, a_size = shell.as_tuple()
                    grid_vals[i, a_ind : a_ind + a_size] = cart_gto3d.cart_gto3d[(La,)](
                        ax, da, A, R
                    )
        return grid_vals @ precontr

    def get_1el_ints_cart(
        self,
        func_dict,
        other=None,
        can_reorder=True,
        components=0,
        screen_func=None,
        **kwargs,
    ):
        shells_a = self
        symmetric = other is None
        shells_b = shells_a if symmetric else other
        cart_bf_num_a = shells_a.cart_bf_num
        cart_bf_num_b = shells_b.cart_bf_num

        # components = 0 indicates, that a plain 2d matrix is desired.
        if is_2d := (components == 0):
            components = 1

        # Preallocate empty matrices and directly assign the calculated values
        integrals = np.zeros((components, cart_bf_num_a, cart_bf_num_b))

        # Dummy screen function that always returns True
        if (not self.screen) or (screen_func is None):
            screen_func = lambda *args: True

        for i, shell_a in enumerate(shells_a):
            La, A, da, ax, a_ind, a_size = shell_a.as_tuple()
            a_slice = slice(a_ind, a_ind + a_size)
            # When we don't deal with symmetric matrices we have to iterate over
            # all other basis functions in shells_b, so we reset i to 0.
            if not symmetric:
                i = 0
            a_min_exp = ax.min()  # Minimum exponent used for screening
            for shell_b in shells_b[i:]:
                Lb, B, db, bx, b_ind, b_size = shell_b.as_tuple()
                b_slice = slice(b_ind, b_ind + b_size)
                # Screen shell pair
                b_min_exp = bx.min()  # Minimum exponent used for screening
                R_ab = np.linalg.norm(A - B)  # Shell center distance
                b_size = get_shell_shape(Lb)[0]
                if not screen_func(a_min_exp, b_min_exp, R_ab):
                    continue
                integrals[:, a_slice, b_slice] = func_dict[(La, Lb)](
                    ax[:, None],
                    da[:, None],
                    A,
                    bx[None, :],
                    db[None, :],
                    B,
                    **kwargs,
                )
                # Fill up the lower tringular part of the matrix in the symmetric
                # case.
                # TODO: this may be (?) better done outside of this loop, after
                # everything is calculated...
                if symmetric:
                    integrals[:, b_slice, a_slice] = np.transpose(
                        integrals[:, a_slice, b_slice], axes=(0, 2, 1)
                    )

        # Return plain 2d array if components is set to 0, i.e., remove first axis.
        if is_2d:
            integrals = np.squeeze(integrals, axis=0)

        # Reordering will be disabled, when spherical integrals are desired. They
        # are reordered outside of this function. Reordering them already here
        # would mess up the results after the 2nd reordering.
        if can_reorder and self.ordering == "native":
            integrals = np.einsum(
                "ij,...jk,kl->...il",
                self.P_cart,
                integrals,
                shells_b.P_cart.T,
                optimize="greedy",
            )
        return integrals

    def get_1el_ints_sph(
        self, int_func=None, int_matrix=None, other=None, **kwargs
    ) -> NDArray:
        if int_matrix is None:
            # Disallow reordering of the Cartesian integrals
            int_matrix = self.get_1el_ints_cart(
                int_func, other=other, can_reorder=False, **kwargs
            )

        shells_b = self if other is None else other
        # Reordering (P) and Cartesian to spherical conversion (C).
        PC_a = self.reorder_c2s_coeffs
        PC_b = shells_b.reorder_c2s_coeffs
        int_matrix_sph = np.einsum(
            "ij,...jk,kl->...il", PC_a, int_matrix, PC_b.T, optimize="greedy"
        )
        return int_matrix_sph

    def get_2c2el_ints_cart(self):
        return self.get_1el_ints_cart(int2c2e3d.int2c2e3d)

    def get_2c2el_ints_sph(self):
        return self.get_1el_ints_sph(int2c2e3d.int2c2e3d)

    def get_3c2el_ints_cart(self, shells_aux):
        """Cartesian 3-center-2-electron integrals.

        DO NOT USE THESE INTEGRALS AS THEY ARE RETURNED FROM THIS METHOD.
        These integrals utilize recurrence relations that are only valid,
        when the resulting Cartesian integrals are transformed into spherical
        integrals.

        Contrary to the general function 'get_1el_ints_cart', that supports
        different 'func_dict' arguments and cross-integrals between two
        different shells this function is less general. This function is
        restricted to '_3center2el_sph' and always uses 'self.shells' as well as
        'self.shells_aux'.
        """
        shells_a = self
        cart_bf_num_a = shells_a.cart_bf_num
        cart_bf_num_aux = shells_aux.cart_bf_num

        integrals = np.zeros((cart_bf_num_a, cart_bf_num_a, cart_bf_num_aux))
        # func_dict = _3center2el3d_sph._3center2el3d_sph
        func_dict = int3c2e3d_sph.int3c2e3d_sph

        for i, shell_a in enumerate(shells_a):
            La, A, da, ax, a_ind, a_size = shell_a.as_tuple()
            a_slice = slice(a_ind, a_ind + a_size)
            # As noted in the docstring, we iterate over pairs of self.shells (shells_a)
            for shell_b in shells_a[i:]:
                Lb, B, db, bx, b_ind, b_size = shell_b.as_tuple()
                b_slice = slice(b_ind, b_ind + b_size)
                for shell_c in shells_aux:
                    Lc, C, dc, cx, c_ind, c_size = shell_c.as_tuple()
                    c_slice = slice(c_ind, c_ind + c_size)
                    integrals[a_slice, b_slice, c_slice] = func_dict[(La, Lb, Lc)](
                        ax[:, None, None],
                        da[:, None, None],
                        A,
                        bx[None, :, None],
                        db[None, :, None],
                        B,
                        cx[None, None, :],
                        dc[None, None, :],
                        C,
                    )
                    integrals[b_slice, a_slice, c_slice] = np.transpose(
                        integrals[a_slice, b_slice, c_slice], axes=(1, 0, 2)
                    )
        return integrals

    def get_3c2el_ints_sph(self, shells_aux):
        int_matrix = self.get_3c2el_ints_cart(shells_aux)

        PC_a = self.reorder_c2s_coeffs
        PC_c = shells_aux.reorder_c2s_coeffs
        int_matrix_sph = np.einsum(
            "ij,jlm,nl,om->ino", PC_a, int_matrix, PC_a, PC_c, optimize="greedy"
        )
        return int_matrix_sph

    #####################
    # Overlap integrals #
    #####################

    @staticmethod
    def screen_S(a, b, R, k=10):
        """
        Returns False when distance R is below calculated threshold.
        Derived from the sympy code below.

        from sympy import *

        a, b, S, R = symbols("a b S R", real=True, positive=True)
        k = symbols("k", integer=True, positive=True)
        apb = a + b
        # 0 = S - 10**(-k)
        ovlp = (pi / apb)**Rational(3,2) * exp(-a*b / apb * R**2) - 10**(-k)
        print("ss-overlap:", ovlp)
        # Distance R, when ss-overlaps drops below 10**(-k)
        R = solve(ovlp, R)[0]
        print("R:", R)

        Parameters
        ----------
        a
            Minimum exponent in shell a.
        b
            Minimum exponent in shell b.
        R
            Real space distance of shells a and b.
        """
        return R < (
            sqrt(a + b)
            * sqrt(log(10**k * pi ** (3 / 2) / (a + b) ** (3 / 2)))
            / (sqrt(a) * sqrt(b))
        )

    def get_S_cart(self, other=None) -> NDArray:
        # return self.get_1el_ints_cart(ovlp, other=other, screen_func=Shells.screen_S)
        return self.get_1el_ints_cart(
            ovlp3d.ovlp3d, other=other, screen_func=Shells.screen_S
        )

    def get_S_sph(self, other=None) -> NDArray:
        return self.get_1el_ints_sph(
            ovlp3d.ovlp3d, other=other, screen_func=Shells.screen_S
        )

    @property
    def S_cart(self):
        return self.get_S_cart()

    @property
    def S_sph(self):
        return self.get_S_sph()

    ############################
    # Kinetic energy integrals #
    ############################

    def get_T_cart(self, other=None) -> NDArray:
        # return self.get_1el_ints_cart(kinetic, other=other)
        return self.get_1el_ints_cart(kinetic3d.kinetic3d, other=other)

    def get_T_sph(self, other=None) -> NDArray:
        return self.get_1el_ints_sph(kinetic3d.kinetic3d, other=other)

    @property
    def T_cart(self):
        return self.get_T_cart()

    @property
    def T_sph(self):
        return self.get_T_sph()

    ################################
    # Nuclear attraction integrals #
    ################################

    def get_V_cart(self, coords3d=None, charges=None, **kwargs):
        """Nuclear attraction integrals.

        Coordinates and charges must be either be omitted or given together.
        Alternatively, this function could also take one array that combines
        coords3c and charges (not yet implemented).
        """
        assert ((coords3d is None) and (charges is None)) or (
            (coords3d is not None) and (charges is not None)
        )
        if coords3d is None:
            atoms, coords3d = self.atoms_coords3d
            charges = nuc_charges_for_atoms(atoms)

        # def boys_1c1el(n_max, ax, A, bx, B, C):
        # ax = ax[:, None]
        # bx = bx[None, :]
        # p = ax + bx
        # P = -(ax * A + bx * B) / p
        # R_PC2 = ((P - C)**2).sum()
        # pR_PC2 = p * R_PC2
        # return np.array([boys.boys(n, pR_PC2) for n in range(n_max+1)])

        cart_bf_num = self.cart_bf_num
        V_nuc = np.zeros((cart_bf_num, cart_bf_num))
        # Loop over all centers and add their contributions
        for R, Z in zip(coords3d, charges):
            # -Z = -1 * Z, because electrons have negative charge.
            V_nuc += -Z * self.get_1el_ints_cart(
                coulomb3d.coulomb3d,
                R=R,
                **kwargs,
            )
        return V_nuc

    def get_V_sph(self, coords3d=None, charges=None) -> NDArray:
        V_cart = self.get_V_cart(coords3d, charges, can_reorder=False)
        return self.get_1el_ints_sph(int_func=None, int_matrix=V_cart)

    @property
    def V_cart(self):
        return self.get_V_cart()

    @property
    def V_sph(self):
        return self.get_V_sph()

    ###########################
    # Dipole moment integrals #
    ###########################

    def get_dipole_ints_cart(self, origin):
        return self.get_1el_ints_cart(
            dipole3d.dipole3d, components=3, R=origin, screen_func=Shells.screen_S
        )

    def get_dipole_ints_sph(self, origin) -> NDArray:
        return self.get_1el_ints_sph(
            dipole3d.dipole3d, components=3, R=origin, screen_func=Shells.screen_S
        )

    ##################################################
    # Quadrupole moment integrals, diagonal elements #
    ##################################################

    def get_diag_quadrupole_ints_cart(self, origin):
        return self.get_1el_ints_cart(
            diag_quadrupole3d.diag_quadrupole3d,
            components=3,
            R=origin,
            screen_func=Shells.screen_S,
        )

    def get_diag_quadrupole_ints_sph(self, origin):
        return self.get_1el_ints_sph(
            diag_quadrupole3d.diag_quadrupole3d,
            components=3,
            R=origin,
            screen_func=Shells.screen_S,
        )

    ###############################
    # Quadrupole moment integrals #
    ###############################

    def get_quadrupole_ints_cart(self, origin):
        ints_flat = self.get_1el_ints_cart(
            quadrupole3d.quadrupole3d,
            components=6,
            R=origin,
            screen_func=Shells.screen_S,
        )
        return multi_component_sym_mat(ints_flat, 3)

    def get_quadrupole_ints_sph(self, origin) -> NDArray:
        ints_flat = self.get_1el_ints_sph(
            quadrupole3d.quadrupole3d,
            components=6,
            R=origin,
            screen_func=Shells.screen_S,
        )
        return multi_component_sym_mat(ints_flat, 3)

    def to_sap_shells(self):
        for shell in self.shells:
            shell.to_sap_shell()

    def __str__(self):
        return f"{self.__class__.__name__}({len(self.shells)} shells, ordering={self.ordering})"


class AOMixShells(Shells):
    cart_order = (
        ("",),
        ("x", "y", "z"),
        ("xx", "yy", "zz", "xy", "xz", "yz"),
        ("xxx", "yyy", "zzz", "xyy", "xxy", "xxz", "xzz", "yzz", "yyz", "xyz"),
        (
            "zzzz",
            "yzzz",
            "yyzz",
            "yyyz",
            "yyyy",
            "xzzz",
            "xyzz",
            "xyyz",
            "xyyy",
            "xxzz",
            "xxyz",
            "xxyy",
            "xxxz",
            "xxxy",
            "xxxx",
        ),
    )


class ORCAShells(Shells):
    sph_Ps = {
        0: [[1]],  # s
        1: [[0, 1, 0], [1, 0, 0], [0, 0, 1]],  # pz px py
        2: [
            [0, 0, 1, 0, 0],  # dz²
            [0, 1, 0, 0, 0],  # dxz
            [0, 0, 0, 1, 0],  # dyz
            [1, 0, 0, 0, 0],  # dx² - y²
            [0, 0, 0, 0, 1],  # dxy
        ],
        3: [
            [0, 0, 0, 1, 0, 0, 0],
            [0, 0, 1, 0, 0, 0, 0],
            [0, 0, 0, 0, 1, 0, 0],
            [0, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 1, 0],
            [-1, 0, 0, 0, 0, 0, 0],  # ORCA, why you do this to me?
            [0, 0, 0, 0, 0, 0, -1],  # sign flip, as in the line above
        ],
        4: [
            [0, 0, 0, 0, 1, 0, 0, 0, 0],
            [0, 0, 0, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 1, 0, 0, 0],
            [0, 0, 1, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 1, 0, 0],
            [0, -1, 0, 0, 0, 0, 0, 0, 0],  # sign flip
            [0, 0, 0, 0, 0, 0, 0, -1, 0],  # sign flip
            [-1, 0, 0, 0, 0, 0, 0, 0, 0],  # sign flip
            [0, 0, 0, 0, 0, 0, 0, 0, -1],  # sign flip
        ],
    }


class FCHKShells(Shells):
    cart_order = (
        ("",),
        ("x", "y", "z"),
        ("xx", "yy", "zz", "xy", "xz", "yz"),
        ("xxx", "yyy", "zzz", "xyy", "xxy", "xxz", "xzz", "yzz", "yyz", "xyz"),
        (
            "zzzz",
            "yzzz",
            "yyzz",
            "yyyz",
            "yyyy",
            "xzzz",
            "xyzz",
            "xyyz",
            "xyyy",
            "xxzz",
            "xxyz",
            "xxyy",
            "xxxz",
            "xxxy",
            "xxxx",
        ),
    )

    sph_Ps = {
        0: [[1]],  # s
        1: [[1, 0, 0], [0, 0, 1], [0, 1, 0]],  # px pz py
        2: [
            [0, 0, 1, 0, 0],  # dz²
            [0, 1, 0, 0, 0],  # dxz
            [0, 0, 0, 1, 0],  # dyz
            [1, 0, 0, 0, 0],  # dx² - y²
            [0, 0, 0, 0, 1],  # dxy
        ],
        # Similar to ORCA, but w/o the sign flips ;)
        3: [
            [0, 0, 0, 1, 0, 0, 0],
            [0, 0, 1, 0, 0, 0, 0],
            [0, 0, 0, 0, 1, 0, 0],
            [0, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 1, 0],
            [1, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 1],
        ],
        4: [
            [0, 0, 0, 0, 1, 0, 0, 0, 0],
            [0, 0, 0, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 1, 0, 0, 0],
            [0, 0, 1, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 1, 0, 0],
            [0, 1, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 1, 0],
            [1, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 1],
        ],
    }


class MoldenShells(Shells):
    sph_Ps = {
        0: [[1]],  # s
        1: [[1, 0, 0], [0, 0, 1], [0, 1, 0]],  # px, py, pz
        2: [
            [0, 0, 1, 0, 0],  # dz²
            [0, 1, 0, 0, 0],  # dxz
            [0, 0, 0, 1, 0],  # dyz
            [1, 0, 0, 0, 0],  # dx² - y²
            [0, 0, 0, 0, 1],  # dxy
        ],
        3: [
            [0, 0, 0, 1, 0, 0, 0],
            [0, 0, 1, 0, 0, 0, 0],
            [0, 0, 0, 0, 1, 0, 0],
            [0, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 1, 0],
            [1, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 1],
        ],
        4: [
            [0, 0, 0, 0, 1, 0, 0, 0, 0],
            [0, 0, 0, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 1, 0, 0, 0],
            [0, 0, 1, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 1, 0, 0],
            [0, 1, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 1, 0],
            [1, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 1],
        ],
    }


class ORCAMoldenShells(Shells):
    sph_Ps = {
        0: [[1]],  # s
        1: [[1, 0, 0], [0, 0, 1], [0, 1, 0]],  # px, py, pz
        2: [
            [0, 0, 1, 0, 0],  # dz²
            [0, 1, 0, 0, 0],  # dxz
            [0, 0, 0, 1, 0],  # dyz
            [1, 0, 0, 0, 0],  # dx² - y²
            [0, 0, 0, 0, 1],  # dxy
        ],
        3: [
            [0, 0, 0, 1, 0, 0, 0],
            [0, 0, 1, 0, 0, 0, 0],
            [0, 0, 0, 0, 1, 0, 0],
            [0, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 1, 0],
            [-1, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, -1],
        ],
        4: [
            [0, 0, 0, 0, 1, 0, 0, 0, 0],
            [0, 0, 0, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 1, 0, 0, 0],
            [0, 0, 1, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 1, 0, 0],
            [0, -1, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, -1, 0],
            [-1, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, -1],
        ],
    }


def pyscf_cart_order(l):
    order = list()
    for lx in reversed(range(l + 1)):
        for ly in reversed(range(l + 1 - lx)):
            lz = l - lx - ly
            order.append("x" * lx + "y" * ly + "z" * lz)
    return tuple(order)


class PySCFShells(Shells):

    """
    Cartesian bfs >= d angular momentum are not normalized!
    S_ref = mol.intor("int1e_ovlp_cart")
    N = 1 / np.diag(S_ref)**0.5
    ao *= N
    """

    cart_order = [pyscf_cart_order(l) for l in range(5)]

    sph_Ps = {
        0: [[1]],  # s
        1: [[1, 0, 0], [0, 0, 1], [0, 1, 0]],  # px py pz
        2: [
            [0, 0, 0, 0, 1],  # dxy
            [0, 0, 0, 1, 0],  # dyz
            [0, 0, 1, 0, 0],  # dz²
            [0, 1, 0, 0, 0],  # dxz
            [1, 0, 0, 0, 0],  # dx² - y²
        ],
        3: [
            [0, 0, 0, 0, 0, 0, 1],
            [0, 0, 0, 0, 0, 1, 0],
            [0, 0, 0, 0, 1, 0, 0],
            [0, 0, 0, 1, 0, 0, 0],
            [0, 0, 1, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0, 0],
            [1, 0, 0, 0, 0, 0, 0],
        ],
        4: [
            [0, 0, 0, 0, 0, 0, 0, 0, 1],
            [0, 0, 0, 0, 0, 0, 0, 1, 0],
            [0, 0, 0, 0, 0, 0, 1, 0, 0],
            [0, 0, 0, 0, 0, 1, 0, 0, 0],
            [0, 0, 0, 0, 1, 0, 0, 0, 0],
            [0, 0, 0, 1, 0, 0, 0, 0, 0],
            [0, 0, 1, 0, 0, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0, 0, 0, 0],
            [1, 0, 0, 0, 0, 0, 0, 0, 0],
        ],
    }



from pysisyphus.calculators.AnaPotBase import AnaPotBase

# [1] https://aip.scitation.org/doi/abs/10.1063/1.442352
#     Cerjan, Miller


class CerjanMiller(AnaPotBase):

    def __init__(self, a=1, b=1, c=1): 
        # Eq. (3.1) in [1]
        V_str = f"({a}-{b}*y**2)*x**2*exp(-x**2)+{c}/2*y**2"
        xlim = (-1.3, 1.3)
        ylim = (-0.7, 1.9)
        super().__init__(V_str=V_str, xlim=xlim, ylim=ylim)

    def __str__(self):
        return "CerjanMiller calculator"


if __name__ == "__main__":
    cj = CerjanMiller()
    cj.plot()
    import matplotlib.pyplot as plt
    plt.show()
